"""Exceptions raised by the clouddirectory service."""

import json

from moto.core.exceptions import JsonRESTError


class ValidationError(JsonRESTError):
    def __init__(self, message: str):
        super().__init__("ValidationException", message)


class InvalidArnException(JsonRESTError):
    def __init__(self, resource_id: str):
        super().__init__("InvalidArnException", "Invalid Arn")
        body = {
            "ResourceId": resource_id,
            "Message": "Invalid Arn",
        }
        self.description = json.dumps(body)


class ResourceNotFoundException(JsonRESTError):
    def __init__(self, resource_id: str):
        super().__init__("ResourceNotFoundException", "Resource not found")
        body = {
            "ResourceId": resource_id,
            "Message": "Resource not found",
        }
        self.description = json.dumps(body)


class SchemaAlreadyPublishedException(JsonRESTError):
    def __init__(self, schema_arn: str):
        super().__init__("SchemaAlreadyPublishedException", "Schema already published")
        body = {
            "SchemaArn": schema_arn,
            "Message": "Schema already published",
        }
        self.description = json.dumps(body)


class SchemaAlreadyExistsException(JsonRESTError):
    def __init__(self, schema_arn: str):
        super().__init__("SchemaAlreadyExistsException", "Schema already exists")
        body = {
            "SchemaArn": schema_arn,
            "Message": "Schema already exists",
        }
        self.description = json.dumps(body)



from typing import Any, Final

from moto.stepfunctions.parser.api import (
    ExecutionFailedEventDetails,
    HistoryEventType,
)
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
    FailureEventException,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name import (
    StatesErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name_type import (
    StatesErrorNameType,
)
from moto.stepfunctions.parser.asl.component.common.string.string_expression import (
    StringSampler,
)
from moto.stepfunctions.parser.asl.component.state.wait.wait_function.wait_function import (
    WaitFunction,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails
from moto.stepfunctions.parser.asl.utils.json_path import NoSuchJsonPathError


class SecondsPath(WaitFunction):
    # SecondsPath
    # A time, in seconds, to state_wait before beginning the state specified in the Next
    # field, specified using a path from the state's input data.
    # You must specify an integer value for this field.
    string_sampler: Final[StringSampler]

    def __init__(self, string_sampler: StringSampler):
        self.string_sampler = string_sampler

    def _validate_seconds_value(self, env: Environment, seconds: Any):
        if isinstance(seconds, int) and seconds >= 0:
            return
        error_type = StatesErrorNameType.StatesRuntime

        assignment_description = f"{self.string_sampler.literal_value} == {seconds}"
        if not isinstance(seconds, int):
            cause = f"The SecondsPath parameter cannot be parsed as a long value: {assignment_description}"
        else:  # seconds < 0
            cause = f"The SecondsPath parameter references a negative value: {assignment_description}"

        raise FailureEventException(
            failure_event=FailureEvent(
                env=env,
                error_name=StatesErrorName(typ=error_type),
                event_type=HistoryEventType.ExecutionFailed,
                event_details=EventDetails(
                    executionFailedEventDetails=ExecutionFailedEventDetails(
                        error=error_type.to_name(), cause=cause
                    )
                ),
            )
        )

    def _get_wait_seconds(self, env: Environment) -> int:
        try:
            self.string_sampler.eval(env=env)
        except NoSuchJsonPathError as no_such_json_path_error:
            cause = f"The SecondsPath parameter does not reference an input value: {no_such_json_path_error.json_path}"
            raise FailureEventException(
                failure_event=FailureEvent(
                    env=env,
                    error_name=StatesErrorName(typ=StatesErrorNameType.StatesRuntime),
                    event_type=HistoryEventType.ExecutionFailed,
                    event_details=EventDetails(
                        executionFailedEventDetails=ExecutionFailedEventDetails(
                            error=StatesErrorNameType.StatesRuntime.to_name(),
                            cause=cause,
                        )
                    ),
                )
            )
        seconds = env.stack.pop()
        self._validate_seconds_value(env=env, seconds=seconds)
        return seconds



import re

from boto3 import Session

from moto.eks.exceptions import InvalidParameterException


def validate_role_arn(arn: str) -> None:
    valid_role_arn_format = re.compile(
        "arn:(?P<partition>.+):iam::(?P<account_id>[0-9]{12}):role/.+"
    )
    match = valid_role_arn_format.match(arn)
    valid_partition = match.group("partition") in Session().get_available_partitions()  # type: ignore

    if not all({arn, match, valid_partition}):
        raise InvalidParameterException("Invalid Role Arn: '" + arn + "'")  # type: ignore



"""
This module provides a facade over the `botocore.model` module.

It defines custom shape and service model classes that extend Botocore's
classes to add additional functionality and properties specific to Moto.

Because Botocore's shape classes are not designed for subclassing, this module
redefines the class hierarchy to ensure that all shape classes inherit from
both the relevant Botocore shape class as well as a new Moto base class.
This allows us to add common functionality to all shape types while still
retaining the original behavior of the Botocore classes without monkey-patching.

This module also defines a modified ShapeResolver to ensure that all shapes created
through the resolver are instances of our custom classes.

Any Moto code that needs to interact with Botocore's model classes should do so
through this facade to avoid direct dependencies on Botocore's internal structure,
which may change over time.
"""

from __future__ import annotations

from collections.abc import Mapping
from typing import Any, cast

from botocore.model import ListShape as BotocoreListShape
from botocore.model import MapShape as BotocoreMapShape
from botocore.model import OperationModel as BotocoreOperationModel
from botocore.model import ServiceModel as BotocoreServiceModel
from botocore.model import Shape as BotocoreShape
from botocore.model import ShapeResolver as BotocoreShapeResolver
from botocore.model import StringShape as BotocoreStringShape
from botocore.model import StructureShape as BotocoreStructureShape
from botocore.utils import CachedProperty, instance_cache


class Shape(BotocoreShape):
    # Custom Moto model properties that we want available in the serialization dict.
    SERIALIZED_ATTRS = BotocoreShape.SERIALIZED_ATTRS + [
        "locationNameForQueryCompatibility"
    ]

    serialization: dict[str, Any]

    @property
    def is_flattened(self) -> bool:
        return self.serialization.get("flattened", False)

    @property
    def is_http_header_trait(self) -> bool:
        return self.serialization.get("location") in ["header", "headers"]

    @property
    def is_not_bound_to_body(self) -> bool:
        return "location" in self.serialization


class StringShape(BotocoreStringShape, Shape):
    pass


class ListShape(BotocoreListShape, Shape):
    pass


class MapShape(BotocoreMapShape, Shape):
    pass


class StructureShape(BotocoreStructureShape, Shape):
    @CachedProperty
    def members(self) -> dict[str, Shape]:  # type: ignore[override]
        return cast(dict[str, Shape], super().members)

    @CachedProperty
    def error_code_aliases(self) -> list[str]:
        if not self.metadata.get("exception", False):
            return []
        error_metadata = self.metadata.get("error", {})
        aliases = error_metadata.get("codeAliases", [])
        return aliases


class ServiceModel(BotocoreServiceModel):
    def __init__(
        self, service_description: Mapping[str, Any], service_name: str | None = None
    ):
        super().__init__(service_description, service_name)
        # Use our custom shape resolver.
        self._shape_resolver = ShapeResolver(service_description.get("shapes", {}))

    @instance_cache
    def operation_model(self, operation_name: str) -> OperationModel:  # type: ignore[misc]
        operation_model = super().operation_model(operation_name)
        model = getattr(operation_model, "_operation_model", {})
        return OperationModel(model, self, operation_name)

    @CachedProperty
    def is_query_compatible(self) -> bool:
        return "awsQueryCompatible" in self.metadata


class OperationModel(BotocoreOperationModel):
    _operation_model: dict[str, Any]
    _service_model: ServiceModel

    @property
    def service_model(self) -> ServiceModel:
        return self._service_model


class ShapeResolver(BotocoreShapeResolver):
    SHAPE_CLASSES = {
        "structure": StructureShape,  # type: ignore[dict-item]
        "list": ListShape,  # type: ignore[dict-item]
        "map": MapShape,  # type: ignore[dict-item]
        "string": StringShape,  # type: ignore[dict-item]
    }

    def get_shape_by_name(
        self, shape_name: str, member_traits: Mapping[str, Any] | None = None
    ) -> Shape:
        shape = super().get_shape_by_name(shape_name, member_traits)
        # The SHAPE_CLASSES dict only allows us to override Shape subclasses,
        # but Botocore will return its own Shape base class for unknown types.
        # If the returned shape is not a part of our hierarchy, we need to
        # convert it to our Shape base class here.
        if not isinstance(shape, Shape):
            shape = Shape(shape_name, getattr(shape, "_shape_model", {}), self)
        return shape



from typing import Final

from moto.stepfunctions.parser.asl.component.common.string.string_expression import (
    StringSampler,
)
from moto.stepfunctions.parser.asl.component.eval_component import EvalComponent
from moto.stepfunctions.parser.asl.eval.environment import Environment


class ItemsPath(EvalComponent):
    string_sampler: Final[StringSampler]

    def __init__(self, string_sampler: StringSampler):
        self.string_sampler = string_sampler

    def _eval_body(self, env: Environment) -> None:
        self.string_sampler.eval(env=env)



from moto.core.responses import ActionResult, EmptyResult
from moto.core.utils import camelcase_to_underscores
from moto.ec2.utils import add_tag_specification

from ..models.managed_prefixes import ManagedPrefixList
from ._base_response import EC2BaseResponse


class VPCs(EC2BaseResponse):
    def create_default_vpc(self) -> ActionResult:
        vpc = self.ec2_backend.create_default_vpc()
        result = {"Vpc": vpc}
        return ActionResult(result)

    def create_vpc(self) -> ActionResult:
        cidr_block = self._get_param("CidrBlock")
        tags = self._get_param("TagSpecifications", [])
        instance_tenancy = self._get_param("InstanceTenancy", if_none="default")
        amazon_provided_ipv6_cidr_block = self._get_param(
            "AmazonProvidedIpv6CidrBlock", False
        )
        ipv6_cidr_block_network_border_group = self._get_param(
            "Ipv6CidrBlockNetworkBorderGroup"
        )
        # if network group is not specified, use the region of the VPC
        if not ipv6_cidr_block_network_border_group:
            ipv6_cidr_block_network_border_group = self.region
        if tags:
            tags = tags[0].get("Tags")
        vpc = self.ec2_backend.create_vpc(
            cidr_block,
            instance_tenancy,
            amazon_provided_ipv6_cidr_block=amazon_provided_ipv6_cidr_block,
            ipv6_cidr_block_network_border_group=ipv6_cidr_block_network_border_group,
            tags=tags,
        )
        result = {"Vpc": vpc}
        return ActionResult(result)

    def delete_vpc(self) -> ActionResult:
        vpc_id = self._get_param("VpcId")
        self.ec2_backend.delete_vpc(vpc_id)
        return EmptyResult()

    def describe_vpcs(self) -> ActionResult:
        self.error_on_dryrun()
        vpc_ids = self._get_param("VpcIds", [])
        filters = self._filters_from_querystring()
        vpcs = self.ec2_backend.describe_vpcs(vpc_ids=vpc_ids, filters=filters)
        result = {"Vpcs": vpcs}
        return ActionResult(result)

    def modify_vpc_tenancy(self) -> ActionResult:
        vpc_id = self._get_param("VpcId")
        tenancy = self._get_param("InstanceTenancy")
        self.ec2_backend.modify_vpc_tenancy(vpc_id, tenancy)
        return EmptyResult()

    def describe_vpc_attribute(self) -> ActionResult:
        vpc_id = self._get_param("VpcId")
        attribute = self._get_param("Attribute")
        attr_name = camelcase_to_underscores(attribute)
        value = self.ec2_backend.describe_vpc_attribute(vpc_id, attr_name)
        result = {"VpcId": vpc_id, attribute: {"Value": value}}
        return ActionResult(result)

    def describe_vpc_classic_link_dns_support(self) -> ActionResult:
        vpc_ids = self._get_param("VpcIds")
        filters = self._filters_from_querystring()
        vpcs = self.ec2_backend.describe_vpcs(vpc_ids=vpc_ids, filters=filters)
        result = {
            "Vpcs": [
                {
                    "VpcId": vpc.id,
                    "ClassicLinkDnsSupported": vpc.classic_link_dns_supported,
                }
                for vpc in vpcs
            ]
        }
        return ActionResult(result)

    def enable_vpc_classic_link_dns_support(self) -> ActionResult:
        vpc_id = self._get_param("VpcId")
        ret = self.ec2_backend.enable_vpc_classic_link_dns_support(vpc_id=vpc_id)
        return ActionResult({"Return": ret})

    def disable_vpc_classic_link_dns_support(self) -> ActionResult:
        vpc_id = self._get_param("VpcId")
        ret = self.ec2_backend.disable_vpc_classic_link_dns_support(vpc_id=vpc_id)
        return ActionResult({"Return": ret})

    def describe_vpc_classic_link(self) -> ActionResult:
        vpc_ids = self._get_param("VpcIds")
        filters = self._filters_from_querystring()
        vpcs = self.ec2_backend.describe_vpcs(vpc_ids=vpc_ids, filters=filters)
        result = {
            "Vpcs": [
                {
                    "VpcId": vpc.id,
                    "ClassicLinkEnabled": vpc.classic_link_enabled,
                    "Tags": vpc.tag_list,
                }
                for vpc in vpcs
            ]
        }
        return ActionResult(result)

    def enable_vpc_classic_link(self) -> ActionResult:
        vpc_id = self._get_param("VpcId")
        ret = self.ec2_backend.enable_vpc_classic_link(vpc_id=vpc_id)
        return ActionResult({"Return": ret})

    def disable_vpc_classic_link(self) -> ActionResult:
        vpc_id = self._get_param("VpcId")
        ret = self.ec2_backend.disable_vpc_classic_link(vpc_id=vpc_id)
        return ActionResult({"Return": ret})

    def modify_vpc_attribute(self) -> ActionResult:
        vpc_id = self._get_param("VpcId")
        for attribute in (
            "EnableDnsSupport",
            "EnableDnsHostnames",
            "EnableNetworkAddressUsageMetrics",
        ):
            if self._get_param(f"{attribute}.Value") is not None:
                attr_name = camelcase_to_underscores(attribute)
                attr_value = self._get_param(f"{attribute}.Value")
                self.ec2_backend.modify_vpc_attribute(vpc_id, attr_name, attr_value)
        return EmptyResult()

    def associate_vpc_cidr_block(self) -> ActionResult:
        vpc_id = self._get_param("VpcId")
        amazon_provided_ipv6_cidr_blocks = self._get_param(
            "AmazonProvidedIpv6CidrBlock"
        )
        # todo test on AWS if can create an association for IPV4 and IPV6 in the same call?
        cidr_block = (
            self._get_param("CidrBlock")
            if not amazon_provided_ipv6_cidr_blocks
            else None
        )
        association = self.ec2_backend.associate_vpc_cidr_block(
            vpc_id, cidr_block, amazon_provided_ipv6_cidr_blocks
        )
        result = {"VpcId": vpc_id}
        if not amazon_provided_ipv6_cidr_blocks:
            result["CidrBlockAssociation"] = {
                "AssociationId": association["association_id"],
                "CidrBlock": association["cidr_block"],
                "CidrBlockState": {"State": "associating"},
            }
        else:
            result["Ipv6CidrBlockAssociation"] = {
                "AssociationId": association["association_id"],
                "Ipv6CidrBlock": association["cidr_block"],
                "Ipv6CidrBlockState": {"State": "associating"},
                "NetworkBorderGroup": association.get(
                    "ipv6_cidr_block_network_border_group"
                ),
                "Ipv6Pool": association.get("ipv6_pool"),
            }
        return ActionResult(result)

    def disassociate_vpc_cidr_block(self) -> ActionResult:
        association_id = self._get_param("AssociationId")
        association = self.ec2_backend.disassociate_vpc_cidr_block(association_id)
        result = {"VpcId": association["vpc_id"]}
        if "::" in association.get("cidr_block", ""):
            result["Ipv6CidrBlockAssociation"] = {
                "AssociationId": association["association_id"],
                "Ipv6CidrBlock": association["cidr_block"],
                "Ipv6CidrBlockState": {"State": "disassociating"},
                "NetworkBorderGroup": association.get(
                    "ipv6_cidr_block_network_border_group"
                ),
                "Ipv6Pool": association.get("ipv6_pool"),
            }
        else:
            result["CidrBlockAssociation"] = {
                "AssociationId": association["association_id"],
                "CidrBlock": association["cidr_block"],
                "CidrBlockState": {"State": "disassociating"},
            }
        return ActionResult(result)

    def create_vpc_endpoint(self) -> ActionResult:
        vpc_id = self._get_param("VpcId")
        service_name = self._get_param("ServiceName")
        route_table_ids = self._get_param("RouteTableIds", [])
        subnet_ids = self._get_param("SubnetIds", [])
        endpoint_type = self._get_param("VpcEndpointType")
        policy_document = self._get_param("PolicyDocument")
        client_token = self._get_param("ClientToken")
        private_dns_enabled = self._get_bool_param("PrivateDnsEnabled", if_none=True)
        security_group_ids = self._get_param("SecurityGroupIds", [])

        tags = add_tag_specification(self._get_param("TagSpecifications", []))
        vpc_end_point = self.ec2_backend.create_vpc_endpoint(
            vpc_id=vpc_id,
            service_name=service_name,
            endpoint_type=endpoint_type,
            policy_document=policy_document,
            route_table_ids=route_table_ids,
            subnet_ids=subnet_ids,
            client_token=client_token,
            security_group_ids=security_group_ids,
            tags=tags,
            private_dns_enabled=private_dns_enabled,
        )
        result = {"VpcEndpoint": vpc_end_point}
        return ActionResult(result)

    def modify_vpc_endpoint(self) -> ActionResult:
        vpc_id = self._get_param("VpcEndpointId")
        add_subnets = self._get_param("AddSubnetIds", [])
        remove_subnets = self._get_param("RemoveSubnetIds", [])
        add_route_tables = self._get_param("AddRouteTableIds", [])
        remove_route_tables = self._get_param("RemoveRouteTableIds", [])
        policy_doc = self._get_param("PolicyDocument")
        add_security_groups = self._get_param("AddSecurityGroupIds", [])
        remove_security_groups = self._get_param("RemoveSecurityGroupIds", [])
        self.ec2_backend.modify_vpc_endpoint(
            vpc_id=vpc_id,
            policy_doc=policy_doc,
            add_subnets=add_subnets,
            remove_subnets=remove_subnets,
            add_route_tables=add_route_tables,
            remove_route_tables=remove_route_tables,
            add_security_groups=add_security_groups,
            remove_security_groups=remove_security_groups,
        )
        return ActionResult({"Return": True})

    def describe_vpc_endpoint_services(self) -> ActionResult:
        vpc_end_point_services = self.ec2_backend.describe_vpc_endpoint_services(
            service_names=self._get_param("ServiceNames", []),
            filters=self._get_param("Filters", []),
            max_results=self._get_int_param("MaxResults"),
            next_token=self._get_param("NextToken"),
            region=self.region,
        )
        result = {
            "ServiceNames": vpc_end_point_services["serviceNames"],
            "ServiceDetails": vpc_end_point_services["servicesDetails"],
            "NextToken": vpc_end_point_services["nextToken"],
        }
        return ActionResult(result)

    def describe_vpc_endpoints(self) -> ActionResult:
        vpc_end_points_ids = self._get_param("VpcEndpointIds", [])
        filters = self._filters_from_querystring()
        vpc_end_points = self.ec2_backend.describe_vpc_endpoints(
            vpc_end_point_ids=vpc_end_points_ids, filters=filters
        )
        result = {"VpcEndpoints": vpc_end_points}
        return ActionResult(result)

    def delete_vpc_endpoints(self) -> ActionResult:
        vpc_end_points_ids = self._get_param("VpcEndpointIds", [])
        self.ec2_backend.delete_vpc_endpoints(vpce_ids=vpc_end_points_ids)
        return ActionResult({"Unsuccessful": []})

    def create_managed_prefix_list(self) -> ActionResult:
        address_family = self._get_param("AddressFamily")
        max_entries = self._get_param("MaxEntries")
        prefix_list_name = self._get_param("PrefixListName")
        entry = self._get_param("Entries", [])

        tags = self._parse_tag_specification().get("prefix-list", {})

        managed_prefix_list = self.ec2_backend.create_managed_prefix_list(
            address_family=address_family,
            entry=entry,
            max_entries=max_entries,
            prefix_list_name=prefix_list_name,
            tags=tags,
        )
        result = {"PrefixList": managed_prefix_list}
        return ActionResult(result)

    def describe_managed_prefix_lists(self) -> ActionResult:
        prefix_list_ids = self._get_param("PrefixListIds", [])
        filters = self._filters_from_querystring()
        managed_prefix_lists = self.ec2_backend.describe_managed_prefix_lists(
            prefix_list_ids=prefix_list_ids, filters=filters
        )
        result = {"PrefixLists": managed_prefix_lists}
        return ActionResult(result)

    def get_managed_prefix_list_entries(self) -> ActionResult:
        prefix_list_id = self._get_param("PrefixListId")
        target_version = self._get_param("TargetVersion")
        managed_prefix_list = self.ec2_backend.get_managed_prefix_list_entries(
            prefix_list_id=prefix_list_id
        )
        entries: list[ManagedPrefixList] = []
        if managed_prefix_list:
            entries = (
                list(managed_prefix_list.entries.values())[-1]
                if managed_prefix_list.entries.values()
                else []
            )
            if target_version:
                target_version = int(target_version)
                entries = managed_prefix_list.entries.get(target_version)
        result = {"Entries": entries}
        return ActionResult(result)

    def delete_managed_prefix_list(self) -> ActionResult:
        prefix_list_id = self._get_param("PrefixListId")
        managed_prefix_list = self.ec2_backend.delete_managed_prefix_list(
            prefix_list_id
        )
        result = {"PrefixList": managed_prefix_list}
        return ActionResult(result)

    def describe_prefix_lists(self) -> ActionResult:
        prefix_list_ids = self._get_param("PrefixListIds", [])
        filters = self._filters_from_querystring()
        managed_pls = self.ec2_backend.describe_managed_prefix_lists(
            prefix_list_ids=prefix_list_ids, filters=filters
        )
        result = {
            "PrefixLists": [
                pl
                for pl in managed_pls
                if pl.prefix_list_name.startswith("com.amazonaws.")
            ]
        }
        return ActionResult(result)

    def modify_managed_prefix_list(self) -> ActionResult:
        add_entry = self._get_param("AddEntries", [])
        prefix_list_id = self._get_param("PrefixListId")
        current_version = self._get_param("CurrentVersion")
        prefix_list_name = self._get_param("PrefixListName")
        remove_entry = self._get_param("RemoveEntries", [])

        current_version = int(current_version) if current_version else None

        managed_prefix_list = self.ec2_backend.modify_managed_prefix_list(
            add_entry=add_entry,
            prefix_list_id=prefix_list_id,
            current_version=current_version,
            prefix_list_name=prefix_list_name,
            remove_entry=remove_entry,
        )
        result = {"PrefixList": managed_prefix_list}
        return ActionResult(result)



from __future__ import annotations

from typing import TYPE_CHECKING
from urllib.parse import urlparse

from werkzeug.wrappers import Request

from moto.core.utils import gzip_decompress
from moto.settings import MAX_FORM_MEMORY_SIZE
from moto.utilities.constants import APPLICATION_JSON, JSON_TYPES

if TYPE_CHECKING:
    from botocore.awsrequest import AWSPreparedRequest

    from moto.core.model import ServiceModel


def normalize_request(request: AWSPreparedRequest | Request) -> Request:
    if isinstance(request, Request):
        return request
    body = request.body
    # Request.from_values() does not automatically handle gzip-encoded bodies,
    # like the full WSGI server would, so we need to do it manually.
    if request.headers.get("Content-Encoding") == "gzip":
        body = gzip_decompress(body)  # type: ignore[arg-type]
    parsed_url = urlparse(request.url)
    Request.max_form_memory_size = MAX_FORM_MEMORY_SIZE
    normalized_request = Request.from_values(
        method=request.method,
        base_url=f"{parsed_url.scheme}://{parsed_url.netloc}",
        path=parsed_url.path,
        query_string=parsed_url.query,
        data=body,
        headers=[(k, v) for k, v in request.headers.items()],
    )
    return normalized_request


def determine_request_protocol(
    service_model: ServiceModel, content_type: str | None = None
) -> str:
    protocol = str(service_model.protocol)
    supported_protocols = service_model.metadata.get("protocols", [protocol])
    content_type = content_type if content_type is not None else ""
    if content_type in JSON_TYPES:
        protocol = "rest-json" if content_type == APPLICATION_JSON else "json"
    elif content_type.startswith("application/x-www-form-urlencoded"):
        protocol = "ec2" if "ec2" in supported_protocols else "query"
    if protocol not in supported_protocols:
        raise NotImplementedError(
            f"Unsupported protocol [{protocol}] for service {service_model.service_name}"
        )
    return protocol



import json
from json import JSONDecodeError
from typing import IO, Any, Final, Optional, Union

from moto.stepfunctions.parser.api import InvocationResponse
from moto.stepfunctions.parser.asl.component.state.exec.state_task.credentials import (
    StateCredentials,
)
from moto.stepfunctions.parser.asl.component.state.exec.state_task.mock_eval_utils import (
    eval_mocked_response,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.utils.boto_client import boto_client_for
from moto.stepfunctions.parser.asl.utils.encoding import to_json_str
from moto.stepfunctions.parser.mocking.mock_config import MockedResponse
from moto.utilities.collections import select_from_typed_dict


class LambdaFunctionErrorException(Exception):
    function_error: Final[Optional[str]]
    payload: Final[str]

    def __init__(self, function_error: Optional[str], payload: str):
        self.function_error = function_error
        self.payload = payload


def _from_payload(payload_streaming_body: IO[bytes]) -> Union[dict, str]:
    """
    This method extracts the lambda payload. The payload may be a string or a JSON stringified object.
    In the first case, this function converts the output into a UTF-8 string, otherwise it parses the
    JSON string into a JSON object.
    """

    payload_bytes: bytes = payload_streaming_body.read()
    decoded_data: str = payload_bytes.decode("utf-8")
    try:
        json_data: json = json.loads(decoded_data)
        return json_data
    except (UnicodeDecodeError, json.JSONDecodeError):
        return decoded_data


def _mocked_invoke_lambda_function(env: Environment):
    mocked_response: MockedResponse = env.get_current_mocked_response()
    eval_mocked_response(env=env, mocked_response=mocked_response)
    invocation_resp = env.stack.pop()
    return invocation_resp


def _invoke_lambda_function(
    parameters: dict, region: str, state_credentials: StateCredentials
):
    lambda_client = boto_client_for(
        service="lambda", region=region, state_credentials=state_credentials
    )

    invocation_response = lambda_client.invoke(**parameters)

    payload = invocation_response["Payload"]
    payload_json = _from_payload(payload)
    invocation_response["Payload"] = payload_json

    return invocation_response


def exec_lambda_function(
    env: Environment,
    parameters: dict,
    region: str,
    account: str,
    credentials: Any,
) -> None:
    lambda_client = boto_client_for(
        region=region, service="lambda", state_credentials=credentials
    )

    invocation_resp = lambda_client.invoke(**parameters)

    func_error: Optional[str] = invocation_resp.get("FunctionError")
    payload_json = json.load(invocation_resp["Payload"])
    if func_error:
        payload_str = json.dumps(payload_json, separators=(",", ":"))
        raise LambdaFunctionErrorException(func_error, payload_str)

    invocation_resp["Payload"] = payload_json

    env.stack.append(invocation_resp)


def execute_lambda_function_integration(
    env: Environment, parameters: dict, region: str, state_credentials: StateCredentials
) -> None:
    if env.is_mocked_mode():
        invocation_response = _mocked_invoke_lambda_function(env=env)
    else:
        invocation_response = _invoke_lambda_function(
            parameters=parameters, region=region, state_credentials=state_credentials
        )

    function_error: Optional[str] = invocation_response.get("FunctionError")
    if function_error:
        payload_json = invocation_response["Payload"]
        payload_str = json.dumps(payload_json, separators=(",", ":"))
        raise LambdaFunctionErrorException(function_error, payload_str)

    response = select_from_typed_dict(
        typed_dict=InvocationResponse, obj=invocation_response
    )  # noqa
    env.stack.append(response)


def to_payload_type(payload: Any) -> Optional[bytes]:
    if isinstance(payload, bytes):
        return payload

    if payload is None:
        str_value = to_json_str({})
    elif isinstance(payload, str):
        try:
            json.loads(payload)
            str_value = payload
        except JSONDecodeError:
            str_value = to_json_str(payload)
    else:
        str_value = to_json_str(payload)
    return str_value.encode("utf-8")



"""Handles incoming connect requests, invokes methods, returns responses."""

import json
from typing import Any, Optional
from urllib.parse import unquote

from moto.core.responses import BaseResponse

from .models import ConnectBackend, connect_backends


class ConnectResponse(BaseResponse):
    """Handler for Connect requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="connect")

    @property
    def connect_backend(self) -> ConnectBackend:
        """Return backend instance specific for this region."""
        return connect_backends[self.current_account][self.region]

    def _get_instance_id(self) -> str:
        """Extract instance_id from request path params."""
        instance_id = self._get_param("InstanceId")
        return unquote(instance_id) if instance_id else ""

    def associate_analytics_data_set(self) -> str:
        instance_id = self._get_instance_id()
        params = json.loads(self.body) if self.body else {}
        data_set_id = str(params["DataSetId"])
        target_account_id = params.get("TargetAccountId")

        result = self.connect_backend.associate_analytics_data_set(
            instance_id=instance_id,
            data_set_id=data_set_id,
            target_account_id=target_account_id,
        )

        return json.dumps(result)

    def disassociate_analytics_data_set(self) -> str:
        instance_id = self._get_instance_id()
        params = json.loads(self.body) if self.body else {}
        data_set_id = str(params["DataSetId"])

        self.connect_backend.disassociate_analytics_data_set(
            instance_id=instance_id,
            data_set_id=data_set_id,
        )

        return "{}"

    def list_analytics_data_associations(self) -> str:
        instance_id = self._get_instance_id()
        data_set_id = self._get_param("DataSetId")
        max_results = self._get_int_param("maxResults")
        next_token = self._get_param("nextToken")

        results: list[dict[str, str]]
        token: Optional[str]
        results, token = self.connect_backend.list_analytics_data_associations(
            instance_id=instance_id,
            data_set_id=data_set_id,
            max_results=max_results,
            next_token=next_token,
        )

        response: dict[str, Any] = {"Results": results}
        if token:
            response["NextToken"] = token

        return json.dumps(response)

    def create_instance(self) -> str:
        params = json.loads(self.body) if self.body else {}
        identity_management_type = str(params["IdentityManagementType"])
        instance_alias = params.get("InstanceAlias")
        inbound_calls_enabled = params.get("InboundCallsEnabled", False)
        outbound_calls_enabled = params.get("OutboundCallsEnabled", False)
        tags = params.get("Tags")

        result = self.connect_backend.create_instance(
            identity_management_type=identity_management_type,
            instance_alias=instance_alias,
            inbound_calls_enabled=inbound_calls_enabled,
            outbound_calls_enabled=outbound_calls_enabled,
            tags=tags,
        )

        return json.dumps(result)

    def describe_instance(self) -> str:
        instance_id = self._get_instance_id()

        instance = self.connect_backend.describe_instance(instance_id=instance_id)

        return json.dumps({"Instance": instance})

    def list_instances(self) -> str:
        max_results = self._get_int_param("maxResults")
        next_token = self._get_param("nextToken")

        results: list[dict[str, Any]]
        token: Optional[str]
        results, token = self.connect_backend.list_instances(
            max_results=max_results,
            next_token=next_token,
        )

        response: dict[str, Any] = {"InstanceSummaryList": results}
        if token:
            response["NextToken"] = token

        return json.dumps(response)

    def delete_instance(self) -> str:
        instance_id = self._get_instance_id()

        self.connect_backend.delete_instance(instance_id=instance_id)

        return "{}"

    def _get_resource_arn(self) -> str:
        """Extract resourceArn from request path params."""
        resource_arn = self._get_param("resourceArn")
        return unquote(resource_arn) if resource_arn else ""

    def tag_resource(self) -> str:
        resource_arn = self._get_resource_arn()
        params = json.loads(self.body) if self.body else {}
        tags = params.get("tags", {})

        self.connect_backend.tag_resource(resource_arn=resource_arn, tags=tags)

        return "{}"

    def untag_resource(self) -> str:
        resource_arn = self._get_resource_arn()
        tag_keys = self.querystring.get("tagKeys", [])

        self.connect_backend.untag_resource(
            resource_arn=resource_arn, tag_keys=tag_keys
        )

        return "{}"

    def list_tags_for_resource(self) -> str:
        resource_arn = self._get_resource_arn()

        tags = self.connect_backend.list_tags_for_resource(resource_arn=resource_arn)

        return json.dumps({"tags": tags})



import datetime
import re
from typing import Final, Optional

from moto.stepfunctions.parser.api import ExecutionFailedEventDetails, HistoryEventType
from moto.stepfunctions.parser.asl.component.common.error_name.failure_event import (
    FailureEvent,
    FailureEventException,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name import (
    StatesErrorName,
)
from moto.stepfunctions.parser.asl.component.common.error_name.states_error_name_type import (
    StatesErrorNameType,
)
from moto.stepfunctions.parser.asl.component.common.string.string_expression import (
    StringExpression,
    StringLiteral,
)
from moto.stepfunctions.parser.asl.component.state.wait.wait_function.wait_function import (
    WaitFunction,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment
from moto.stepfunctions.parser.asl.eval.event.event_detail import EventDetails

TIMESTAMP_FORMAT: Final[str] = "%Y-%m-%dT%H:%M:%SZ"
# TODO: could be a bit more exact (e.g. 90 shouldn't be a valid minute)
TIMESTAMP_PATTERN: Final[str] = r"^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.\d+)?Z$"


class Timestamp(WaitFunction):
    string: Final[StringExpression]

    def __init__(self, string: StringExpression):
        self.string = string
        # If a string literal, assert it encodes a valid timestamp.
        if isinstance(string, StringLiteral):
            timestamp = string.literal_value
            if self._from_timestamp_string(timestamp) is None:
                raise ValueError(
                    "The Timestamp value does not reference a valid ISO-8601 "
                    f"extended offset date-time format string: '{timestamp}'"
                )

    @staticmethod
    def _is_valid_timestamp_pattern(timestamp: str) -> bool:
        return re.match(TIMESTAMP_PATTERN, timestamp) is not None

    @staticmethod
    def _from_timestamp_string(timestamp: str) -> Optional[datetime.datetime]:
        if not Timestamp._is_valid_timestamp_pattern(timestamp):
            return None
        try:
            # anything lower than seconds is truncated
            processed_timestamp = timestamp.rsplit(".", 2)[0]
            # add back the "Z" suffix if we removed it
            if not processed_timestamp.endswith("Z"):
                processed_timestamp = f"{processed_timestamp}Z"
            datetime_timestamp = datetime.datetime.strptime(
                processed_timestamp, TIMESTAMP_FORMAT
            )
            return datetime_timestamp
        except Exception:
            return None

    def _create_failure_event(
        self, env: Environment, timestamp_str: str
    ) -> FailureEvent:
        return FailureEvent(
            env=env,
            error_name=StatesErrorName(typ=StatesErrorNameType.StatesRuntime),
            event_type=HistoryEventType.ExecutionFailed,
            event_details=EventDetails(
                executionFailedEventDetails=ExecutionFailedEventDetails(
                    error=StatesErrorNameType.StatesRuntime.to_name(),
                    cause="The Timestamp parameter does not reference a valid ISO-8601 "
                    f"extended offset date-time format string: {self.string.literal_value} == {timestamp_str}",
                )
            ),
        )

    def _get_wait_seconds(self, env: Environment) -> int:
        self.string.eval(env=env)
        timestamp_str: str = env.stack.pop()
        timestamp = self._from_timestamp_string(timestamp=timestamp_str)
        if timestamp is None:
            raise FailureEventException(self._create_failure_event(env, timestamp_str))
        delta = timestamp - datetime.datetime.now()
        delta_sec = int(delta.total_seconds())
        return delta_sec


class TimestampPath(Timestamp):
    def _create_failure_event(
        self, env: Environment, timestamp_str: str
    ) -> FailureEvent:
        return FailureEvent(
            env=env,
            error_name=StatesErrorName(typ=StatesErrorNameType.StatesRuntime),
            event_type=HistoryEventType.ExecutionFailed,
            event_details=EventDetails(
                executionFailedEventDetails=ExecutionFailedEventDetails(
                    error=StatesErrorNameType.StatesRuntime.to_name(),
                    cause="The TimestampPath parameter does not reference a valid ISO-8601 "
                    f"extended offset date-time format string: {self.string.literal_value} == {timestamp_str}",
                )
            ),
        )



# mypy: ignore-errors
import base64
import json
from collections import OrderedDict
from collections.abc import MutableMapping
from datetime import datetime, timezone
from typing import Any, Optional

from botocore import xform_name
from botocore.utils import parse_timestamp as botocore_parse_timestamp

from moto.core.model import OperationModel

UNDEFINED = object()  # Sentinel to signal the absence of a field in the input


def parse_timestamp(value: str) -> datetime:
    """Parse a timestamp and return a naive datetime object in UTC.
    This matches Moto's internal representation of timestamps, based
    on moto.core.utils.utcnow().
    """
    parsed = botocore_parse_timestamp(value)
    as_naive_utc = parsed.astimezone(timezone.utc).replace(tzinfo=None)
    return as_naive_utc


def default_blob_parser(value):
    # We don't decode this to a str because it's possible that
    # the blob contains binary data that can't be decoded.
    return base64.b64decode(value)


class RequestParser:
    MAP_TYPE = dict

    def __init__(
        self,
        timestamp_parser=None,
        blob_parser=None,
        map_type=None,
    ):
        if timestamp_parser is None:
            timestamp_parser = parse_timestamp
        self._timestamp_parser = timestamp_parser
        if blob_parser is None:
            blob_parser = default_blob_parser
        self._blob_parser = blob_parser
        if map_type is not None:
            self.MAP_TYPE = map_type

    def parse(self, request_dict, operation_model: OperationModel):
        raise NotImplementedError()


class QueryParser(RequestParser):
    def parse(self, request_dict, operation_model: OperationModel):
        shape = operation_model.input_shape
        if shape is None:
            return {}
        parsed = self._do_parse(request_dict, shape)
        return parsed if parsed is not UNDEFINED else {}

    def _do_parse(self, request_dict, shape):
        parsed = self._parse_shape(shape, request_dict["query_params"])
        return parsed if parsed is not UNDEFINED else {}

    def _parse_shape(self, shape, node, prefix=""):
        handler = getattr(self, f"_handle_{shape.type_name}", self._default_handle)
        return handler(shape, node, prefix)

    def _gonna_recurse(self, query_params, prefix):
        if prefix == "":
            return False
        return not any(param_key.startswith(prefix) for param_key in query_params)

    def _handle_structure(self, shape, query_params, prefix=""):
        if self._gonna_recurse(query_params, prefix):
            return UNDEFINED
        parsed = self.MAP_TYPE()
        members = shape.members
        for member_name in members:
            member_shape = members[member_name]
            member_prefix = self._get_serialized_name(member_shape, member_name)
            if prefix:
                member_prefix = f"{prefix}.{member_prefix}"
            value = self._parse_shape(member_shape, query_params, member_prefix)
            parsed_key = self._parsed_key_name(member_name)
            if value is not UNDEFINED:
                parsed[parsed_key] = value
        return parsed if parsed != {} else UNDEFINED

    def _handle_list(self, shape, node, prefix=""):
        # The query protocol serializes empty lists as an empty string.
        if node.get(prefix, UNDEFINED) == "":
            return []

        if shape.is_flattened:
            list_prefix = prefix
            location_name = self._get_serialized_name(shape.member, None)
            if location_name is not None:
                list_prefix = ".".join(prefix.split(".")[:-1] + [location_name])
        else:
            list_name = self._get_serialized_name(shape.member, "member")
            list_prefix = f"{prefix}.{list_name}"
        parsed_list = []
        i = 1
        while True:
            element_name = f"{list_prefix}.{i}"
            element_shape = shape.member
            value = self._parse_shape(element_shape, node, element_name)
            if value is UNDEFINED:
                break
            parsed_list.append(value)
            i += 1
        return parsed_list if parsed_list != [] else UNDEFINED

    def _handle_map(self, shape, query_params, prefix=""):
        if shape.is_flattened:
            full_prefix = prefix
        else:
            full_prefix = f"{prefix}.entry"
        template = full_prefix + ".{i}.{suffix}"
        key_shape = shape.key
        value_shape = shape.value
        key_suffix = self._get_serialized_name(key_shape, default_name="key")
        value_suffix = self._get_serialized_name(value_shape, "value")
        parsed_map = self.MAP_TYPE()
        i = 1
        while True:
            key_name = template.format(i=i, suffix=key_suffix)
            value_name = template.format(i=i, suffix=value_suffix)
            key = self._parse_shape(key_shape, query_params, key_name)
            value = self._parse_shape(value_shape, query_params, value_name)
            if key is UNDEFINED:
                break
            parsed_map[key] = value
            i += 1
        return parsed_map if parsed_map != {} else UNDEFINED

    def _handle_timestamp(self, shape, query_params, prefix=""):
        value = self._default_handle(shape, query_params, prefix)
        return value if value is UNDEFINED else self._timestamp_parser(value)

    def _handle_blob(self, shape, query_params, prefix=""):
        value = self._default_handle(shape, query_params, prefix)
        if value is UNDEFINED:
            return value
        return self._blob_parser(value)

    def _handle_boolean(self, shape, query_params, prefix=""):
        value = self._default_handle(shape, query_params, prefix)
        try:
            return value.lower() == "true"
        except AttributeError:
            pass
        return UNDEFINED

    def _handle_integer(self, shape, query_params, prefix=""):
        value = self._default_handle(shape, query_params, prefix)
        return value if value is UNDEFINED else int(value)

    def _handle_float(self, shape, query_params, prefix=""):
        value = self._default_handle(shape, query_params, prefix)
        return value if value is UNDEFINED else float(value)

    _handle_double = _handle_float
    _handle_long = _handle_integer

    def _default_handle(self, shape, value, prefix=""):
        default_value = shape.metadata.get("default", UNDEFINED)
        return value.get(prefix, default_value)

    def _get_serialized_name(self, shape, default_name):
        serialized_name = shape.serialization.get("locationNameForQueryCompatibility")
        if serialized_name:
            return serialized_name
        return shape.serialization.get("name", default_name)

    def _parsed_key_name(self, member_name):
        key_name = member_name
        return key_name


class EC2QueryParser(QueryParser):
    def _get_serialized_name(self, shape, default_name):
        # https://smithy.io/2.0/aws/protocols/aws-ec2-query-protocol.html#query-key-resolution
        if "queryName" in shape.serialization:
            return shape.serialization["queryName"]
        if "name" in shape.serialization:
            name = shape.serialization["name"]
            return name[0].upper() + name[1:]
        return default_name

    def _handle_list(self, shape, node, prefix=""):
        parsed_list = []
        i = 1
        while True:
            element_name = f"{prefix}.{i}"
            element_shape = shape.member
            value = self._parse_shape(element_shape, node, element_name)
            if value is UNDEFINED:
                break
            parsed_list.append(value)
            i += 1
        return parsed_list if parsed_list != [] else UNDEFINED


class JSONParser(RequestParser):
    DEFAULT_ENCODING = "utf-8"

    def parse(self, request_dict, operation_model: OperationModel):
        shape = operation_model.input_shape
        parsed = self._do_parse(request_dict, shape)
        return parsed if parsed is not UNDEFINED else {}

    def _do_parse(self, request_dict, shape):
        parsed = self.MAP_TYPE()
        if shape is not None:
            parsed = self._handle_json_body(request_dict["body"], shape)
        return parsed

    def _handle_json_body(self, raw_body, shape):
        parsed_json = self._parse_body_as_json(raw_body)
        return self._parse_shape(shape, parsed_json)

    def _parse_body_as_json(self, body_contents):
        if not body_contents:
            return {}
        body = body_contents.decode(self.DEFAULT_ENCODING)
        original_parsed = json.loads(body)
        return original_parsed

    def _parse_shape(self, shape, node):
        handler = getattr(self, f"_handle_{shape.type_name}", self._default_handle)
        return handler(shape, node)

    def _default_handle(self, _, value):
        return value

    def _handle_blob(self, shape, value):
        value = self._default_handle(shape, value)
        if value is UNDEFINED:
            return value
        return self._blob_parser(value)

    def _handle_float(self, _, value):
        return float(value) if value is not UNDEFINED else value

    def _handle_list(self, shape, node):
        parsed = []
        member_shape = shape.member
        for item in node:
            parsed.append(self._parse_shape(member_shape, item))
        return parsed

    def _handle_map(self, shape, value):
        parsed = self.MAP_TYPE()
        key_shape = shape.key
        value_shape = shape.value
        for key, item_value in value.items():
            actual_key = self._parse_shape(key_shape, key)
            actual_value = self._parse_shape(value_shape, item_value)
            parsed[actual_key] = actual_value
        return parsed

    def _handle_structure(self, shape, value):
        member_shapes = shape.members
        final_parsed = self.MAP_TYPE()
        for member_name in member_shapes:
            member_shape = member_shapes[member_name]
            json_name = member_shape.serialization.get("name", member_name)
            raw_value = value.get(json_name)
            if raw_value is not None:
                final_parsed[member_name] = self._parse_shape(
                    member_shapes[member_name], raw_value
                )
        return final_parsed

    def _handle_timestamp(self, _, value):
        return self._timestamp_parser(value)

    _handle_double = _handle_float


PROTOCOL_PARSERS = {
    "ec2": EC2QueryParser,
    "json": JSONParser,
    "query": QueryParser,
}


class XFormedDict(MutableMapping[str, Any]):
    """
    A Pascal/Snake case-insensitive  ``dict``-like object.

        xfd = XFormedDict()
        xfd['DBInstanceIdentifier'] = 'identifier'
        xfd['DBInstanceIdentifier'] == 'identifier'  # True
        xfd['db_instance_identifier'] == 'identifier'  # True
        list(xfd) == ['db_instance_identifier']  # True

    ***Do Not Import*** This class is deprecated and will be removed in a future release.
    """

    def __init__(
        self, data: Optional[dict[str, Any]] = None, **kwargs: dict[str, Any]
    ) -> None:
        self._xform_cache: dict[str, Any] = {}
        self._store: MutableMapping[str, Any] = OrderedDict()
        if data is None:
            data = {}
        self.update(data, **kwargs)

    def _xformed(self, key: str) -> str:
        return xform_name(key, _xform_cache=self._xform_cache)

    def __setitem__(self, key: str, value: Any) -> None:
        # Use the xformed key for lookups, but store the actual key alongside the value.
        self._store[self._xformed(key)] = (key, value)

    def __getitem__(self, key: str) -> Any:
        return self._store[self._xformed(key)][1]

    def __delitem__(self, key: str) -> None:
        del self._store[self._xformed(key)]

    def __iter__(self) -> Any:
        return (key for key in self._store.keys())

    def __len__(self) -> int:
        return len(self._store)

    def original_dict(self) -> dict[str, Any]:
        original_dict = {}
        for _, keyval in self._store.items():
            key = keyval[0]
            value = keyval[1]
            if isinstance(value, XFormedDict):
                value = value.original_dict()
            if isinstance(value, list):
                value = [
                    v.original_dict() if isinstance(v, XFormedDict) else v
                    for v in value
                ]
            original_dict[key] = value
        return original_dict

    def __eq__(self, other: object) -> bool:
        if isinstance(other, dict):
            other = XFormedDict(other)
        else:
            return NotImplemented
        return dict(self.items()) == dict(other.items())



import abc
import operator
import re
from datetime import date, datetime
from enum import Enum
from itertools import repeat
from typing import Any, Optional, Union

from pyparsing import (
    CaselessKeyword,
    Forward,
    OpAssoc,
    ParserElement,
    ParseResults,
    QuotedString,
    Suppress,
    Word,
    alphanums,
    exceptions,
    infix_notation,
    one_of,
    pyparsing_common,
)

try:
    # TODO import directly when depending on pyparsing>=3.1.0
    from pyparsing import DelimitedList
except ImportError:
    # delimited_list is deprecated in favor of DelimitedList in pyparsing 3.1.0
    from pyparsing import delimited_list as DelimitedList  # type: ignore[assignment]

from .exceptions import (
    InvalidFilterFieldNameException,
    InvalidFilterOperatorException,
    InvalidInputException,
    InvalidLogicalOperatorException,
    InvalidStateException,
)


class Logical(str, Enum):
    AND = "AND"
    ANY = "ANY"


class ConditionState(str, Enum):
    SUCCEEDED = "SUCCEEDED"
    FAILED = "FAILED"
    STOPPED = "STOPPED"
    TIMEOUT = "TIMEOUT"
    CANCELLED = "CANCELLED"


class CrawlState(str, Enum):
    RUNNING = "RUNNING"
    CANCELLING = "CANCELLING"
    CANCELLED = "CANCELLED"
    SUCCEEDED = "SUCCEEDED"
    FAILED = "FAILED"
    ERROR = "ERROR"


class Condition:
    def __init__(self, inputs: dict[str, str], index: int) -> None:
        # if there's a job name present, the API creates a job trigger, even if crawler trigger args are provided
        self.job_name = inputs.get("JobName")
        self.crawler_name = inputs.get("CrawlerName")
        self.state: Optional[ConditionState] = None
        self.crawl_state: Optional[CrawlState] = None
        if self.job_name:
            self.crawler_name = None
            if inputs.get("State"):
                self.state = ConditionState(inputs.get("State"))
            else:
                raise InvalidInputException(
                    "CreateTrigger",
                    "State cannot be null or empty, choose from one of the following job state: [SUCCEEDED, STOPPED, FAILED, TIMEOUT]",
                )
        elif self.crawler_name:
            self.job_name = None
            if inputs.get("CrawlState"):
                self.crawl_state = CrawlState(inputs.get("CrawlState"))
            else:
                raise InvalidInputException(
                    "CreateTrigger",
                    "State cannot be null or empty, choose from one of the following crawler state: [SUCCEEDED, FAILED, CANCELLED]",
                )
        else:
            raise InvalidInputException(
                "CreateTrigger",
                "JobName or CrawlerName in condition cannot be null or empty",
            )

        logical_operator = inputs.get("LogicalOperator")
        if logical_operator:
            self.logical_operator = logical_operator
        else:
            raise InvalidInputException(
                "CreateTrigger",
                "Logical operator cannot be null or empty",
            )
        if self.logical_operator != "EQUALS":
            raise InvalidLogicalOperatorException(self.logical_operator, index + 1)

    def as_dict(self) -> dict[str, str]:
        return_dict = {
            "LogicalOperator": self.logical_operator,
        }
        if self.crawler_name:
            return_dict["CrawlerName"] = self.crawler_name
        if self.job_name:
            return_dict["JobName"] = self.job_name
        if self.crawl_state:
            return_dict["CrawlState"] = self.crawl_state.value
        if self.state:
            return_dict["State"] = self.state.value
        return return_dict


class Predicate:
    def __init__(self, inputs: dict[str, Any]) -> None:
        self.logical = Logical(inputs.get("Logical"))

        self.conditions: list[Condition] = [
            Condition(condition, index)
            for index, condition in enumerate(inputs.get("Conditions", []))
        ]

    def as_dict(self) -> dict[str, Union[str, list[dict[str, str]]]]:
        return_dict: dict[str, Union[str, list[dict[str, str]]]] = {}
        if self.logical:
            return_dict["Logical"] = self.logical.value
        if self.conditions:
            return_dict["Conditions"] = [
                condition.as_dict() for condition in self.conditions
            ]
        return return_dict


class Action:
    def __init__(self, inputs: dict[str, Union[str, dict[str, str], int]]) -> None:
        self.arguments = inputs.get("Arguments")
        self.crawler_name = inputs.get("CrawlerName")
        self.job_name = inputs.get("JobName")
        self.notification_property = inputs.get("NotificationProperty")
        self.security_configuration = inputs.get("SecurityConfiguration")
        self.timeout = inputs.get("Timeout")

        if self.crawler_name and self.job_name:
            raise InvalidInputException(
                "CreateTrigger",
                "Both JobName and CrawlerName cannot be set together in an action",
            )

    def as_dict(self) -> dict[str, Union[str, dict[str, str], int]]:
        return_dict: dict[str, Union[str, dict[str, str], int]] = {}
        if self.arguments:
            return_dict["Arguments"] = self.arguments
        if self.crawler_name:
            return_dict["CrawlerName"] = self.crawler_name
        if self.job_name:
            return_dict["JobName"] = self.job_name
        if self.notification_property:
            return_dict["NotificationProperty"] = self.notification_property
        if self.security_configuration:
            return_dict["SecurityConfiguration"] = self.security_configuration
        if self.timeout is not None:
            return_dict["Timeout"] = self.timeout
        return return_dict


class FilterField(str, Enum):
    CRAWL_ID = "CRAWL_ID"
    STATE = "STATE"
    START_TIME = "START_TIME"
    END_TIME = "END_TIME"
    DPU_HOUR = "DPU_HOUR"


class FilterOperator(str, Enum):
    GT = "GT"
    GE = "GE"
    LT = "LT"
    LE = "LE"
    EQ = "EQ"
    NE = "NE"


class CrawlFilter:
    def __init__(
        self, field_name: FilterField, operator: FilterOperator, field_value: str
    ) -> None:
        self.field_name = field_name
        self.field_value = field_value
        self.operator = operator
        self.check_operator_valid_for_state()

    def check_operator_valid_for_state(self) -> None:
        if self.field_name in [FilterField.STATE, FilterField.CRAWL_ID]:
            if self.operator not in [FilterOperator.EQ, FilterOperator.NE]:
                raise InvalidInputException(
                    "ListCrawls",
                    f"Only EQ or NE operator is allowed for field : {self.field_name.value}",
                )


def validate_crawl_filters(filters: list[dict[str, str]]) -> list[CrawlFilter]:
    filter_objects: list[CrawlFilter] = []
    for index, filter in enumerate(filters):
        field_name_input = filter.get("FieldName")
        operator_input = filter.get("FilterOperator")
        field_value_input = filter.get("FieldValue")
        if not all([field_name_input, field_value_input, operator_input]):
            raise InvalidInputException(
                "ListCrawls",
                f"Invalid Filter Provided: FieldName: {field_name_input if field_name_input else 'null'}, "
                f"FilterOperator: {operator_input if operator_input else 'null'}, "
                f"FieldValue: {field_value_input if field_value_input else 'null'}",
            )

        try:
            field_name = FilterField(field_name_input)
        except ValueError:
            raise InvalidFilterFieldNameException(filter["FieldName"], index + 1)

        try:
            operator = FilterOperator(operator_input)
        except ValueError:
            raise InvalidFilterOperatorException(filter["FilterOperator"], index + 1)

        filter_objects.append(CrawlFilter(field_name, operator, filter["FieldValue"]))

    return filter_objects


def _cast(type_: str, value: Any) -> Union[date, datetime, float, int, str]:
    # values are always cast from string to target type
    value = str(value)

    if type_ in ("bigint", "int", "smallint", "tinyint"):
        try:
            return int(value)  # no size is enforced
        except ValueError:
            raise ValueError(f'"{value}" is not an integer.')

    if type_ == "decimal":
        try:
            return float(value)
        except ValueError:
            raise ValueError(f"{value} is not a decimal.")

    if type_ in ("char", "string", "varchar"):
        return value  # no length is enforced

    if type_ == "date":
        try:
            return datetime.strptime(value, "%Y-%m-%d").date()
        except ValueError:
            raise ValueError(f"{value} is not a date.")

    if type_ == "timestamp":
        match = re.search(
            r"^(?P<timestamp>\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2})"
            r"(?P<nanos>\.\d{1,9})?$",
            value,
        )
        if match is None:
            raise ValueError(
                "Timestamp format must be yyyy-mm-dd hh:mm:ss[.fffffffff]"
                f" {value} is not a timestamp."
            )

        try:
            timestamp = datetime.strptime(match.group("timestamp"), "%Y-%m-%d %H:%M:%S")
        except ValueError:
            raise ValueError(
                "Timestamp format must be yyyy-mm-dd hh:mm:ss[.fffffffff]"
                f" {value} is not a timestamp."
            )

        # use nanosecond representation for timestamps
        posix_nanoseconds = int(timestamp.timestamp() * 1_000_000_000)

        nanos = match.group("nanos")
        if nanos is not None:
            # strip leading dot, reverse and left pad with zeros to nanoseconds
            nanos = "".join(reversed(nanos[1:])).zfill(9)
            for i, nanoseconds in enumerate(nanos):
                posix_nanoseconds += int(nanoseconds) * 10**i

        return posix_nanoseconds

    raise InvalidInputException("GetPartitions", f"Unknown type : '{type_}'")


def _escape_regex(pattern: str) -> str:
    """Taken from Python 3.7 to avoid escaping '%'."""
    _special_chars_map = {i: "\\" + chr(i) for i in b"()[]{}?*+-|^$\\.&~# \t\n\r\v\f"}
    return pattern.translate(_special_chars_map)


class _Expr(abc.ABC):
    @abc.abstractmethod
    def eval(self, part_keys: list[dict[str, str]], part_input: dict[str, Any]) -> Any:  # type: ignore[misc]
        raise NotImplementedError()


class _Ident(_Expr):
    def __init__(self, tokens: ParseResults):
        self.ident: str = tokens[0]

    def eval(self, part_keys: list[dict[str, str]], part_input: dict[str, Any]) -> Any:
        try:
            return self._eval(part_keys, part_input)
        except ValueError as e:
            # existing partition values cannot be cast to current schema
            raise InvalidStateException("GetPartitions", str(e))

    def leval(self, part_keys: list[dict[str, str]], literal: Any) -> Any:
        # evaluate literal by simulating partition input
        try:
            return self._eval(part_keys, part_input={"Values": repeat(literal)})
        except ValueError as e:
            # expression literal cannot be cast to current schema
            raise InvalidInputException("GetPartitions", str(e))

    def type_(self, part_keys: list[dict[str, str]]) -> str:
        for key in part_keys:
            if self.ident == key["Name"]:
                return key["Type"]

        # not a partition column
        raise InvalidInputException("GetPartitions", "Invalid partition expression!")

    def _eval(self, part_keys: list[dict[str, str]], part_input: dict[str, Any]) -> Any:
        for key, value in zip(part_keys, part_input["Values"]):
            if self.ident == key["Name"]:
                return _cast(key["Type"], value)

        # also raised for unpartitioned tables
        raise InvalidInputException("GetPartitions", "Invalid partition expression!")


class _IsNull(_Expr):
    def __init__(self, tokens: ParseResults):
        self.ident: _Ident = tokens[0]

    def eval(self, part_keys: list[dict[str, str]], part_input: dict[str, Any]) -> bool:
        return self.ident.eval(part_keys, part_input) is None


class _IsNotNull(_IsNull):
    def eval(self, part_keys: list[dict[str, str]], part_input: dict[str, Any]) -> bool:
        return not super().eval(part_keys, part_input)


class _BinOp(_Expr):
    def __init__(self, tokens: ParseResults):
        self.ident: _Ident = tokens[0]
        self.bin_op: str = tokens[1]
        self.literal: Any = tokens[2]

    def eval(self, part_keys: list[dict[str, str]], part_input: dict[str, Any]) -> bool:
        ident = self.ident.eval(part_keys, part_input)

        # simulate partition input for the lateral
        rhs = self.ident.leval(part_keys, self.literal)

        return {
            "<>": operator.ne,
            ">=": operator.ge,
            "<=": operator.le,
            ">": operator.gt,
            "<": operator.lt,
            "=": operator.eq,
        }[self.bin_op](ident, rhs)


class _Like(_Expr):
    def __init__(self, tokens: ParseResults):
        self.ident: _Ident = tokens[0]
        self.literal: str = tokens[2]

    def eval(self, part_keys: list[dict[str, str]], part_input: dict[str, Any]) -> bool:
        type_ = self.ident.type_(part_keys)
        if type_ in ("bigint", "int", "smallint", "tinyint"):
            raise InvalidInputException(
                "GetPartitions", "Integral data type doesn't support operation 'LIKE'"
            )

        if type_ in ("date", "decimal", "timestamp"):
            raise InvalidInputException(
                "GetPartitions",
                f"{type_[0].upper()}{type_[1:]} data type"
                " doesn't support operation 'LIKE'",
            )

        ident = self.ident.eval(part_keys, part_input)
        assert isinstance(ident, str)

        pattern = _cast("string", self.literal)

        # prepare SQL pattern for conversion to regex pattern
        pattern = _escape_regex(pattern)  # type: ignore

        # NOTE convert SQL wildcards to regex, no literal matches possible
        pattern = pattern.replace("_", ".").replace("%", ".*")

        # LIKE clauses always start at the beginning
        pattern = "^" + pattern + "$"

        return re.search(pattern, ident) is not None


class _NotLike(_Like):
    def eval(self, part_keys: list[dict[str, str]], part_input: dict[str, Any]) -> bool:
        return not super().eval(part_keys, part_input)


class _In(_Expr):
    def __init__(self, tokens: ParseResults):
        self.ident: _Ident = tokens[0]
        self.values: list[Any] = tokens[2:]

    def eval(self, part_keys: list[dict[str, str]], part_input: dict[str, Any]) -> bool:
        ident = self.ident.eval(part_keys, part_input)
        values = (self.ident.leval(part_keys, value) for value in self.values)

        return ident in values


class _NotIn(_In):
    def eval(self, part_keys: list[dict[str, str]], part_input: dict[str, Any]) -> bool:
        return not super().eval(part_keys, part_input)


class _Between(_Expr):
    def __init__(self, tokens: ParseResults):
        self.ident: _Ident = tokens[0]
        self.left: Any = tokens[2]
        self.right: Any = tokens[4]

    def eval(self, part_keys: list[dict[str, str]], part_input: dict[str, Any]) -> bool:
        ident = self.ident.eval(part_keys, part_input)
        left = self.ident.leval(part_keys, self.left)
        right = self.ident.leval(part_keys, self.right)

        return left <= ident <= right or left > ident > right


class _NotBetween(_Between):
    def eval(self, part_keys: list[dict[str, str]], part_input: dict[str, Any]) -> bool:
        return not super().eval(part_keys, part_input)


class _BoolAnd(_Expr):
    def __init__(self, tokens: ParseResults) -> None:
        self.operands: list[_Expr] = tokens[0][0::2]  # skip 'and' between tokens

    def eval(self, part_keys: list[dict[str, str]], part_input: dict[str, Any]) -> bool:
        return all(operand.eval(part_keys, part_input) for operand in self.operands)


class _BoolOr(_Expr):
    def __init__(self, tokens: ParseResults) -> None:
        self.operands: list[_Expr] = tokens[0][0::2]  # skip 'or' between tokens

    def eval(self, part_keys: list[dict[str, str]], part_input: dict[str, Any]) -> bool:
        return any(operand.eval(part_keys, part_input) for operand in self.operands)


class _PartitionFilterExpressionCache:
    def __init__(self) -> None:
        # build grammar according to Glue.Client.get_partitions(Expression)
        lpar, rpar = map(Suppress, "()")

        # NOTE these are AWS Athena column name best practices
        ident = Forward().set_name("ident")
        ident <<= Word(alphanums + "._").set_parse_action(_Ident) | lpar + ident + rpar  # type: ignore

        number = Forward().set_name("number")
        number <<= pyparsing_common.number | lpar + number + rpar  # type: ignore

        string = Forward().set_name("string")
        string <<= QuotedString(quote_char="'", esc_quote="''") | lpar + string + rpar  # type: ignore

        literal = (number | string).set_name("literal")
        literal_list = DelimitedList(literal, min=1).set_name("list")

        bin_op = one_of("<> >= <= > < =").set_name("binary op")

        and_ = Forward()
        and_ <<= CaselessKeyword("and") | lpar + and_ + rpar

        or_ = Forward()
        or_ <<= CaselessKeyword("or") | lpar + or_ + rpar

        in_, between, like, not_, is_, null = map(
            CaselessKeyword, "in between like not is null".split()
        )
        not_ = Suppress(not_)  # type: ignore  # only needed for matching

        cond = (
            (ident + is_ + null).set_parse_action(_IsNull)
            | (ident + is_ + not_ + null).set_parse_action(_IsNotNull)
            | (ident + bin_op + literal).set_parse_action(_BinOp)
            | (ident + like + string).set_parse_action(_Like)
            | (ident + not_ + like + string).set_parse_action(_NotLike)
            | (ident + in_ + lpar + literal_list + rpar).set_parse_action(_In)
            | (ident + not_ + in_ + lpar + literal_list + rpar).set_parse_action(_NotIn)
            | (ident + between + literal + and_ + literal).set_parse_action(_Between)
            | (ident + not_ + between + literal + and_ + literal).set_parse_action(
                _NotBetween
            )
        ).set_name("cond")

        # conditions can be joined using 2-ary AND and/or OR
        expr = infix_notation(
            cond,
            [
                (and_, 2, OpAssoc.LEFT, _BoolAnd),
                (or_, 2, OpAssoc.LEFT, _BoolOr),
            ],
        )
        self._expr = expr.set_name("expr")

        self._cache: dict[str, _Expr] = {}

    def get(self, expression: Optional[str]) -> Optional[_Expr]:
        if expression is None:
            return None

        if expression not in self._cache:
            ParserElement.enable_packrat()

            try:
                expr: ParseResults = self._expr.parse_string(expression, parse_all=True)
                self._cache[expression] = expr[0]
            except exceptions.ParseException:
                raise InvalidInputException(
                    "GetPartitions", f"Unsupported expression '{expression}'"
                )

        return self._cache[expression]


_PARTITION_FILTER_EXPRESSION_CACHE = _PartitionFilterExpressionCache()


class PartitionFilter:
    def __init__(self, expression: Optional[str], fake_table: Any):
        self.expression = expression
        self.fake_table = fake_table

    def __call__(self, fake_partition: Any) -> bool:
        expression = _PARTITION_FILTER_EXPRESSION_CACHE.get(self.expression)
        if expression is None:
            return True

        versions = list(self.fake_table.versions.values())
        return expression.eval(
            part_keys=versions[-1].get("PartitionKeys", []),
            part_input=fake_partition.partition_input,
        )



"""EventBridgePipesBackend class with methods for supported APIs."""

from typing import Any, Optional

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.core.utils import utcnow
from moto.utilities.paginator import paginate
from moto.utilities.utils import get_partition

from .exceptions import NotFoundException

PAGINATION_MODEL = {
    "list_pipes": {
        "input_token": "next_token",
        "limit_key": "limit",
        "limit_default": 100,
        "unique_attribute": "Arn",
    },
}


class Pipe(BaseModel):
    """Represents an EventBridge Pipe."""

    def __init__(
        self,
        name: str,
        account_id: str,
        region_name: str,
        source: str,
        target: str,
        role_arn: str,
        description: Optional[str] = None,
        desired_state: Optional[str] = None,
        source_parameters: Optional[dict[str, Any]] = None,
        enrichment: Optional[str] = None,
        enrichment_parameters: Optional[dict[str, Any]] = None,
        target_parameters: Optional[dict[str, Any]] = None,
        tags: Optional[dict[str, str]] = None,
        log_configuration: Optional[dict[str, Any]] = None,
        kms_key_identifier: Optional[str] = None,
    ):
        self.name = name
        self.account_id = account_id
        self.region_name = region_name
        self.description = description
        self.desired_state = desired_state or "RUNNING"
        self.current_state = "RUNNING"
        self.source = source
        self.source_parameters = source_parameters
        self.enrichment = enrichment
        self.enrichment_parameters = enrichment_parameters
        self.target = target
        self.target_parameters = target_parameters
        self.role_arn = role_arn
        self.tags = tags or {}
        self.log_configuration = log_configuration
        self.kms_key_identifier = kms_key_identifier
        self.creation_time = utcnow()
        self.last_modified_time = utcnow()
        self.state_reason = None

    @property
    def arn(self) -> str:
        partition = get_partition(self.region_name)
        return f"arn:{partition}:pipes:{self.region_name}:{self.account_id}:pipe/{self.name}"


class EventBridgePipesBackend(BaseBackend):
    """Implementation of EventBridgePipes APIs."""

    def __init__(self, region_name: str, account_id: str) -> None:
        super().__init__(region_name, account_id)
        self.pipes: dict[str, Pipe] = {}

    def create_pipe(
        self,
        name: str,
        description: Optional[str],
        desired_state: Optional[str],
        source: str,
        source_parameters: Optional[dict[str, Any]],
        enrichment: Optional[str],
        enrichment_parameters: Optional[dict[str, Any]],
        target: str,
        target_parameters: Optional[dict[str, Any]],
        role_arn: str,
        tags: Optional[dict[str, str]],
        log_configuration: Optional[dict[str, Any]],
        kms_key_identifier: Optional[str],
    ) -> Pipe:
        pipe = Pipe(
            name=name,
            account_id=self.account_id,
            region_name=self.region_name,
            description=description,
            desired_state=desired_state,
            source=source,
            source_parameters=source_parameters,
            enrichment=enrichment,
            enrichment_parameters=enrichment_parameters,
            target=target,
            target_parameters=target_parameters,
            role_arn=role_arn,
            tags=tags,
            log_configuration=log_configuration,
            kms_key_identifier=kms_key_identifier,
        )

        self.pipes[name] = pipe

        return pipe

    def describe_pipe(self, name: str) -> Pipe:
        if name not in self.pipes:
            raise NotFoundException(f"Pipe {name} not found")
        return self.pipes[name]

    def delete_pipe(self, name: str) -> Pipe:
        if name not in self.pipes:
            raise NotFoundException(f"Pipe {name} not found")
        pipe = self.pipes.pop(name)
        pipe.desired_state = "DELETED"
        pipe.current_state = "DELETING"
        return pipe

    def tag_resource(self, resource_arn: str, tags: dict[str, str]) -> None:
        pipe = None
        for p in self.pipes.values():
            if p.arn == resource_arn:
                pipe = p
                break

        if pipe is None:
            raise NotFoundException(f"Resource {resource_arn} not found")

        pipe.tags.update(tags)

    def untag_resource(self, resource_arn: str, tag_keys: list[str]) -> None:
        pipe = None
        for p in self.pipes.values():
            if p.arn == resource_arn:
                pipe = p
                break

        if pipe is None:
            raise NotFoundException(f"Resource {resource_arn} not found")

        for tag_key in tag_keys:
            pipe.tags.pop(tag_key, None)

    def list_tags_for_resource(self, resource_arn: str) -> dict[str, str]:
        pipe = None
        for p in self.pipes.values():
            if p.arn == resource_arn:
                pipe = p
                break

        if pipe is None:
            raise NotFoundException(f"Resource {resource_arn} not found")

        return pipe.tags

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_pipes(  # type: ignore[misc]
        self,
        name_prefix: Optional[str],
        desired_state: Optional[str],
        current_state: Optional[str],
        source_prefix: Optional[str],
        target_prefix: Optional[str],
    ) -> list[dict[str, Any]]:
        """List pipes with optional filtering and pagination."""
        filtered_pipes = list(self.pipes.values())

        if name_prefix:
            filtered_pipes = [
                p for p in filtered_pipes if p.name.startswith(name_prefix)
            ]

        if desired_state:
            filtered_pipes = [
                p for p in filtered_pipes if p.desired_state == desired_state
            ]

        if current_state:
            filtered_pipes = [
                p for p in filtered_pipes if p.current_state == current_state
            ]

        if source_prefix:
            filtered_pipes = [
                p
                for p in filtered_pipes
                if p.source and p.source.startswith(source_prefix)
            ]

        if target_prefix:
            filtered_pipes = [
                p
                for p in filtered_pipes
                if p.target and p.target.startswith(target_prefix)
            ]

        filtered_pipes.sort(key=lambda p: p.name)

        pipe_summaries = []
        for pipe in filtered_pipes:
            summary: dict[str, Any] = {
                "Arn": pipe.arn,
                "Name": pipe.name,
                "DesiredState": pipe.desired_state,
                "CurrentState": pipe.current_state,
                "Source": pipe.source,
                "Target": pipe.target,
                "CreationTime": pipe.creation_time,
                "LastModifiedTime": pipe.last_modified_time,
            }
            if pipe.enrichment:
                summary["Enrichment"] = pipe.enrichment
            pipe_summaries.append(summary)

        return pipe_summaries


pipes_backends = BackendDict(EventBridgePipesBackend, "pipes")



from typing import TYPE_CHECKING, Any, Optional

from moto.dynamodb.exceptions import MockValidationException, ResourceNotFoundException

if TYPE_CHECKING:
    from py_partiql_parser import QueryMetadata


def query(
    statement: str, source_data: dict[str, list[Any]], parameters: list[dict[str, Any]]
) -> tuple[
    list[dict[str, Any]],
    dict[str, list[tuple[Optional[dict[str, Any]], Optional[dict[str, Any]]]]],
]:
    from py_partiql_parser import DynamoDBStatementParser
    from py_partiql_parser.exceptions import DocumentNotFoundException

    try:
        return DynamoDBStatementParser(source_data).parse(statement, parameters)
    except DocumentNotFoundException as dnfe:
        if "." in dnfe.name:
            table_name = dnfe.name.split(".")[0]
            if table_name in source_data:
                raise MockValidationException(
                    message="The table does not have the specified index"
                )
        raise ResourceNotFoundException()


def get_query_metadata(statement: str) -> "QueryMetadata":
    from py_partiql_parser import DynamoDBStatementParser

    return DynamoDBStatementParser.get_query_metadata(query=statement)



from moto.core.responses import ActionResult
from moto.core.utils import utcnow

from ._base_response import EC2BaseResponse


class Windows(EC2BaseResponse):
    def bundle_instance(self) -> str:
        raise NotImplementedError("Windows.bundle_instance is not yet implemented")

    def cancel_bundle_task(self) -> str:
        raise NotImplementedError("Windows.cancel_bundle_task is not yet implemented")

    def describe_bundle_tasks(self) -> str:
        raise NotImplementedError(
            "Windows.describe_bundle_tasks is not yet implemented"
        )

    def get_password_data(self) -> ActionResult:
        instance_id = self._get_param("InstanceId")
        password_data = self.ec2_backend.get_password_data(instance_id)
        result = {
            "InstanceId": instance_id,
            "Timestamp": utcnow(),
            "PasswordData": password_data,
        }
        return ActionResult(result)



from typing import Any, Final, Optional

from moto.stepfunctions.parser.api import Arn, Definition, LongArn, StateMachineType


class AWSExecutionDetails:
    account: Final[str]
    region: Final[str]
    role_arn: Final[str]

    def __init__(self, account: str, region: str, role_arn: str):
        self.account = account
        self.region = region
        self.role_arn = role_arn


class ExecutionDetails:
    arn: Final[LongArn]
    name: Final[str]
    role_arn: Final[Arn]
    inpt: Final[Optional[Any]]
    start_time: Final[str]

    def __init__(
        self,
        arn: LongArn,
        name: str,
        role_arn: Arn,
        inpt: Optional[Any],
        start_time: str,
    ):
        self.arn = arn
        self.name = name
        self.role_arn = role_arn
        self.inpt = inpt
        self.start_time = start_time


class StateMachineDetails:
    arn: Final[Arn]
    name: Final[str]
    typ: Final[StateMachineType]
    definition: Final[Definition]

    def __init__(self, arn: Arn, name: str, typ: StateMachineType, definition: str):
        self.arn = arn
        self.name = name
        self.typ = typ
        self.definition = definition


class EvaluationDetails:
    aws_execution_details: Final[AWSExecutionDetails]
    execution_details: Final[ExecutionDetails]
    state_machine_details: Final[StateMachineDetails]

    def __init__(
        self,
        aws_execution_details: AWSExecutionDetails,
        execution_details: ExecutionDetails,
        state_machine_details: StateMachineDetails,
    ):
        self.aws_execution_details = aws_execution_details
        self.execution_details = execution_details
        self.state_machine_details = state_machine_details



"""Handles incoming sesv2 requests, invokes methods, returns responses."""

import base64
import json

from moto.core.responses import BaseResponse

from .models import SESV2Backend, sesv2_backends


class SESV2Response(BaseResponse):
    """Handler for SESV2 requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="sesv2")

    @property
    def sesv2_backend(self) -> SESV2Backend:
        """Return backend instance specific for this region."""
        return sesv2_backends[self.current_account][self.region]

    def send_email(self) -> str:
        """Piggy back on functionality from v1 mostly"""

        params = json.loads(self.body)
        from_email_address = params.get("FromEmailAddress")
        destination = params.get("Destination", {})
        content = params.get("Content")
        if "Raw" in content:
            all_destinations: list[str] = []
            if "ToAddresses" in destination:
                all_destinations = all_destinations + destination["ToAddresses"]
            if "CcAddresses" in destination:
                all_destinations = all_destinations + destination["CcAddresses"]
            if "BccAddresses" in destination:
                all_destinations = all_destinations + destination["BccAddresses"]
            message = self.sesv2_backend.send_raw_email(
                source=from_email_address,
                destinations=all_destinations,
                raw_data=base64.b64decode(content["Raw"]["Data"]).decode("utf-8"),
            )
        elif "Simple" in content:
            content_body = content["Simple"]["Body"]
            if "Html" in content_body:
                body = content_body["Html"]["Data"]
            else:
                body = content_body["Text"]["Data"]
            message = self.sesv2_backend.send_email(  # type: ignore
                source=from_email_address,
                destinations=destination,
                subject=content["Simple"]["Subject"]["Data"],
                body=body,
            )
        elif "Template" in content:
            raise NotImplementedError("Template functionality not ready")

        return json.dumps({"MessageId": message.id})

    def create_contact_list(self) -> str:
        params = json.loads(self.body)
        self.sesv2_backend.create_contact_list(params)
        return json.dumps({})

    def get_contact_list(self) -> str:
        contact_list_name = self._get_param("ContactListName")
        contact_list = self.sesv2_backend.get_contact_list(contact_list_name)
        return json.dumps(contact_list.response_object)

    def list_contact_lists(self) -> str:
        contact_lists = self.sesv2_backend.list_contact_lists()
        return json.dumps({"ContactLists": [c.response_object for c in contact_lists]})

    def delete_contact_list(self) -> str:
        name = self._get_param("ContactListName")
        self.sesv2_backend.delete_contact_list(name)
        return json.dumps({})

    def create_contact(self) -> str:
        contact_list_name = self._get_param("ContactListName")
        params = json.loads(self.body)
        self.sesv2_backend.create_contact(contact_list_name, params)
        return json.dumps({})

    def get_contact(self) -> str:
        email = self._get_param("EmailAddress")
        contact_list_name = self._get_param("ContactListName")
        contact = self.sesv2_backend.get_contact(email, contact_list_name)
        return json.dumps(contact.response_object)

    def list_contacts(self) -> str:
        contact_list_name = self._get_param("ContactListName")
        contacts = self.sesv2_backend.list_contacts(contact_list_name)
        return json.dumps({"Contacts": [c.response_object for c in contacts]})

    def delete_contact(self) -> str:
        email = self._get_param("EmailAddress")
        contact_list_name = self._get_param("ContactListName")
        self.sesv2_backend.delete_contact(email, contact_list_name)
        return json.dumps({})

    def create_email_identity(self) -> str:
        email_identity_name = self._get_param("EmailIdentity")
        tags = self._get_param("Tags")
        dkim_signing_attributes = self._get_param("DkimSigningAttributes")
        configuration_set_name = self._get_param("ConfigurationSetName")
        email_identity = self.sesv2_backend.create_email_identity(
            email_identity=email_identity_name,
            tags=tags,
            dkim_signing_attributes=dkim_signing_attributes,
            configuration_set_name=configuration_set_name,
        )
        return json.dumps(
            {
                "IdentityType": email_identity.identity_type,
                "VerifiedForSendingStatus": email_identity.verified_for_sending_status,
                "DkimAttributes": email_identity.dkim_attributes,
            }
        )

    def delete_email_identity(self) -> str:
        email_identity_name = self._get_param("EmailIdentity")
        self.sesv2_backend.delete_email_identity(email_identity=email_identity_name)
        return json.dumps({})

    def get_email_identity(self) -> str:
        email_identity_name = self._get_param("EmailIdentity")
        email_identity = self.sesv2_backend.get_email_identity(
            email_identity=email_identity_name,
        )
        return json.dumps(email_identity.get_response_object)

    def list_email_identities(self) -> str:
        next_token = self._get_param("NextToken")
        page_size = self._get_param("PageSize")
        email_identities, next_token = self.sesv2_backend.list_email_identities(
            next_token=next_token,
            page_size=page_size,
        )
        if isinstance(email_identities, list):
            response = [e.list_response_object for e in email_identities]
        else:
            response = []

        return json.dumps(
            {
                "EmailIdentities": response,
                "NextToken": next_token,
            }
        )

    def create_configuration_set(self) -> str:
        configuration_set_name = self._get_param("ConfigurationSetName")
        tracking_options = self._get_param("TrackingOptions")
        delivery_options = self._get_param("DeliveryOptions")
        reputation_options = self._get_param("ReputationOptions")
        sending_options = self._get_param("SendingOptions")
        tags = self._get_param("Tags")
        suppression_options = self._get_param("SuppressionOptions")
        vdm_options = self._get_param("VdmOptions")
        self.sesv2_backend.create_configuration_set(
            configuration_set_name=configuration_set_name,
            tracking_options=tracking_options,
            delivery_options=delivery_options,
            reputation_options=reputation_options,
            sending_options=sending_options,
            tags=tags,
            suppression_options=suppression_options,
            vdm_options=vdm_options,
        )
        return json.dumps({})

    def delete_configuration_set(self) -> str:
        configuration_set_name = self._get_param("ConfigurationSetName")
        self.sesv2_backend.delete_configuration_set(
            configuration_set_name=configuration_set_name,
        )
        return json.dumps({})

    def get_configuration_set(self) -> str:
        configuration_set_name = self._get_param("ConfigurationSetName")
        config_set = self.sesv2_backend.get_configuration_set(
            configuration_set_name=configuration_set_name,
        )
        return json.dumps(config_set.to_dict_v2())

    def list_configuration_sets(self) -> str:
        next_token = self._get_param("NextToken")
        page_size = self._get_param("PageSize")
        configuration_sets, next_token = self.sesv2_backend.list_configuration_sets(
            next_token=next_token, page_size=page_size
        )
        config_set_names = [c.configuration_set_name for c in configuration_sets]

        return json.dumps(
            {"ConfigurationSets": config_set_names, "NextToken": next_token}
        )

    def create_dedicated_ip_pool(self) -> str:
        pool_name = self._get_param("PoolName")
        tags = self._get_param("Tags")
        scaling_mode = self._get_param("ScalingMode")
        self.sesv2_backend.create_dedicated_ip_pool(
            pool_name=pool_name,
            tags=tags,
            scaling_mode=scaling_mode,
        )
        return json.dumps({})

    def delete_dedicated_ip_pool(self) -> str:
        pool_name = self._get_param("PoolName")
        self.sesv2_backend.delete_dedicated_ip_pool(
            pool_name=pool_name,
        )
        return json.dumps({})

    def list_dedicated_ip_pools(self) -> str:
        next_token = self._get_param("NextToken")
        page_size = self._get_param("PageSize")
        dedicated_ip_pools, next_token = self.sesv2_backend.list_dedicated_ip_pools(
            next_token=next_token, page_size=page_size
        )
        return json.dumps(
            {"DedicatedIpPools": dedicated_ip_pools, "NextToken": next_token}
        )

    def get_dedicated_ip_pool(self) -> str:
        pool_name = self._get_param("PoolName")
        dedicated_ip_pool = self.sesv2_backend.get_dedicated_ip_pool(
            pool_name=pool_name,
        )
        return json.dumps({"DedicatedIpPool": dedicated_ip_pool.to_dict()})

    def create_email_identity_policy(self) -> str:
        email_identity = self._get_param("EmailIdentity")
        policy_name = self._get_param("PolicyName")
        policy = self._get_param("Policy")
        self.sesv2_backend.create_email_identity_policy(
            email_identity=email_identity,
            policy_name=policy_name,
            policy=policy,
        )
        return json.dumps({})

    def delete_email_identity_policy(self) -> str:
        email_identity = self._get_param("EmailIdentity")
        policy_name = self._get_param("PolicyName")
        self.sesv2_backend.delete_email_identity_policy(
            email_identity=email_identity,
            policy_name=policy_name,
        )
        return json.dumps({})

    def update_email_identity_policy(self) -> str:
        email_identity = self._get_param("EmailIdentity")
        policy_name = self._get_param("PolicyName")
        policy = self._get_param("Policy")
        self.sesv2_backend.update_email_identity_policy(
            email_identity=email_identity,
            policy_name=policy_name,
            policy=policy,
        )
        return json.dumps({})

    def get_email_identity_policies(self) -> str:
        email_identity = self._get_param("EmailIdentity")
        policies = self.sesv2_backend.get_email_identity_policies(
            email_identity=email_identity,
        )
        return json.dumps({"Policies": policies})

    def tag_resource(self) -> str:
        resource_arn = self._get_param("ResourceArn")
        tags = self._get_param("Tags")
        self.sesv2_backend.tag_resource(
            resource_arn=resource_arn,
            tags=tags,
        )
        return json.dumps({})

    def untag_resource(self) -> str:
        resource_arn = self._get_param("ResourceArn")
        tag_keys = self.__dict__["data"]["TagKeys"]
        self.sesv2_backend.untag_resource(
            resource_arn=resource_arn,
            tag_keys=tag_keys,
        )
        return json.dumps({})

    def list_tags_for_resource(self) -> str:
        resource_arn = self._get_param("ResourceArn")
        tags = self.sesv2_backend.list_tags_for_resource(
            resource_arn=resource_arn,
        )
        return json.dumps({"Tags": tags})



from typing import Any

from moto.core.exceptions import ServiceException


class IAMException(ServiceException):
    pass


class NotFoundException(IAMException):
    code = "NoSuchEntity"


class DeleteConflictException(IAMException):
    code = "DeleteConflict"


class ReportNotPresentException(IAMException):
    code = "ReportNotPresent"


class LimitExceededException(IAMException):
    code = "LimitExceeded"


class MalformedCertificate(IAMException):
    code = "MalformedCertificate"

    def __init__(self, cert: str):
        super().__init__(f"Certificate {cert} is malformed")


class MalformedPolicyDocument(IAMException):
    code = "MalformedPolicyDocument"


class DuplicateTags(IAMException):
    code = "InvalidInput"
    message = (
        "Duplicate tag keys found. Please note that Tag keys are case insensitive."
    )


class ValidationError(IAMException):
    code = "ValidationError"


class TagKeyTooBig(ValidationError):
    def __init__(self, tag: str, param: str = "tags.X.member.key"):
        super().__init__(
            f"1 validation error detected: Value '{tag}' at '{param}' failed to satisfy "
            "constraint: Member must have length less than or equal to 128.",
        )


class TagValueTooBig(ValidationError):
    def __init__(self, tag: str):
        super().__init__(
            f"1 validation error detected: Value '{tag}' at 'tags.X.member.value' failed to satisfy "
            "constraint: Member must have length less than or equal to 256.",
        )


class InvalidTagCharacters(ValidationError):
    def __init__(self, tag: str, param: str = "tags.X.member.key"):
        message = f"1 validation error detected: Value '{tag}' at '{param}' failed to satisfy constraint: Member must satisfy regular expression pattern: [\\p{{L}}\\p{{Z}}\\p{{N}}_.:/=+\\-@]+"
        super().__init__(message)


class TooManyTags(ValidationError):
    def __init__(self, tags: Any, param: str = "tags"):
        super().__init__(
            f"1 validation error detected: Value '{tags}' at '{param}' failed to satisfy "
            "constraint: Member must have length less than or equal to 50.",
        )


class EntityAlreadyExists(IAMException):
    code = "EntityAlreadyExists"


class InvalidInput(IAMException):
    code = "InvalidInput"


class NoSuchEntity(IAMException):
    code = "NoSuchEntity"



"""FirehoseBackend class with methods for supported APIs.

Incomplete list of unfinished items:
  - Data record size and number of transactions are ignored.
  - Better validation of delivery destination parameters, e.g.,
    validation of the url for an http endpoint (boto3 does this).
  - Better handling of the put_record_batch() API.  Not only is
    the existing logic bare bones, but for the ElasticSearch and
    RedShift destinations, the data is just ignored.
  - put_record_batch() handling of errors is minimal and no errors
    are reported back to the user.  Instead an exception is raised.
  - put_record(), put_record_batch() always set "Encrypted" to False.
"""

import io
import json
import warnings
from gzip import GzipFile
from time import time
from typing import Any, Optional

import requests

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.core.types import Base64EncodedString
from moto.core.utils import utcnow
from moto.firehose.exceptions import (
    ConcurrentModificationException,
    InvalidArgumentException,
    LimitExceededException,
    ResourceInUseException,
    ResourceNotFoundException,
    ValidationException,
)
from moto.moto_api._internal import mock_random
from moto.s3.models import s3_backends
from moto.utilities.tagging_service import TaggingService
from moto.utilities.utils import get_partition

MAX_TAGS_PER_DELIVERY_STREAM = 50

DESTINATION_TYPES_TO_NAMES = {
    "s3": "S3",
    "extended_s3": "ExtendedS3",
    "http_endpoint": "HttpEndpoint",
    "elasticsearch": "Elasticsearch",
    "redshift": "Redshift",
    "snowflake": "Snowflake",
    "splunk": "Splunk",  # Unimplemented
}


def find_destination_config_in_args(api_args: dict[str, Any]) -> tuple[str, Any]:
    """Return (config_name, config) tuple for destination config.

    Determines which destination config(s) have been specified.  The
    alternative is to use a bunch of 'if' statements to check each
    destination configuration.  If more than one destination config is
    specified, than an exception is raised.

    A logical name for the destination type is returned along with the
    destination config as it's useful way to compare current and replacement
    destinations.
    """
    destination_names = DESTINATION_TYPES_TO_NAMES.keys()
    configs = []
    for arg_name, arg_value in api_args.items():
        # Ignore arguments that are not destination configs.
        if "_destination" not in arg_name:
            continue

        # If the destination config value is non-null, save it.
        name = arg_name.split("_destination")[0]
        if name in destination_names and arg_value:
            configs.append((DESTINATION_TYPES_TO_NAMES[name], arg_value))

    # One and only one destination configuration is allowed.
    if len(configs) != 1:
        raise InvalidArgumentException(
            "Exactly one destination configuration is supported for a Firehose"
        )

    return configs[0]


def create_s3_destination_config(
    extended_s3_destination_config: dict[str, Any],
) -> dict[str, Any]:
    """Return dict with selected fields copied from ExtendedS3 config.

    When an ExtendedS3 config is chosen, AWS tacks on a S3 config as
    well.  When the same field names for S3 and ExtendedS3 exists,
    the ExtendedS3 fields are copied to the added S3 destination.
    """
    fields_not_needed = [
        "S3BackupMode",
        "S3Description",
        "DataFormatconversionConfiguration",
        "DynamicPartitionConfiguration",
    ]
    destination = {}
    for field, value in extended_s3_destination_config.items():
        if field in fields_not_needed:
            continue
        destination[field] = value
    return destination


class DeliveryStream(BaseModel):
    """Represents a delivery stream, its source and destination configs."""

    STATES = {"CREATING", "ACTIVE", "CREATING_FAILED"}

    MAX_STREAMS_PER_REGION = 50

    ALTERNATIVE_FIELD_NAMES = [
        ("S3Configuration", "S3DestinationDescription"),
        ("S3Update", "S3DestinationDescription"),
        ("S3BackupConfiguration", "S3BackupDescription"),
        ("S3BackupUpdate", "S3BackupDescription"),
    ]

    def __init__(
        self,
        account_id: str,
        region: str,
        delivery_stream_name: str,
        delivery_stream_type: str,
        encryption: dict[str, Any],
        kinesis_stream_source_configuration: dict[str, Any],
        destination_name: str,
        destination_config: dict[str, Any],
    ):
        self.delivery_stream_status = "CREATING"
        self.delivery_stream_name = delivery_stream_name
        self.delivery_stream_type = (
            delivery_stream_type if delivery_stream_type else "DirectPut"
        )
        self.delivery_stream_encryption_configuration = encryption

        self.source = kinesis_stream_source_configuration
        self.destinations: list[dict[str, Any]] = [
            {
                "destination_id": "destinationId-000000000001",
                destination_name: destination_config,
            }
        ]

        if destination_name == "ExtendedS3":
            # Add a S3 destination as well, minus a few ExtendedS3 fields.
            self.destinations[0]["S3"] = create_s3_destination_config(
                destination_config
            )

        # S3Configuration becomes S3DestinationDescription for the
        # other destinations. Same for S3Backup
        for old, new in DeliveryStream.ALTERNATIVE_FIELD_NAMES:
            if old in destination_config:
                self.destinations[0][destination_name][new] = destination_config[old]
                del self.destinations[0][destination_name][old]

        self.delivery_stream_status = "ACTIVE"
        self.delivery_stream_arn = f"arn:{get_partition(region)}:firehose:{region}:{account_id}:deliverystream/{delivery_stream_name}"

        self.create_timestamp = utcnow()
        self.version_id = "1"  # Used to track updates of destination configs

        # I believe boto3 only adds this field after an update ...
        self.last_update_timestamp = utcnow()


class FirehoseBackend(BaseBackend):
    """Implementation of Firehose APIs."""

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.delivery_streams: dict[str, DeliveryStream] = {}
        self.tagger = TaggingService()

    @staticmethod
    def default_vpc_endpoint_service(
        service_region: str, zones: list[str]
    ) -> list[dict[str, str]]:
        """Default VPC endpoint service."""
        return BaseBackend.default_vpc_endpoint_service_factory(
            service_region, zones, "firehose", special_service_name="kinesis-firehose"
        )

    def create_delivery_stream(
        self,
        region: str,
        delivery_stream_name: str,
        delivery_stream_type: str,
        kinesis_stream_source_configuration: dict[str, Any],
        delivery_stream_encryption_configuration_input: dict[str, Any],
        s3_destination_configuration: dict[str, Any],
        extended_s3_destination_configuration: dict[str, Any],
        redshift_destination_configuration: dict[str, Any],
        elasticsearch_destination_configuration: dict[str, Any],
        splunk_destination_configuration: dict[str, Any],
        http_endpoint_destination_configuration: dict[str, Any],
        snowflake_destination_configuration: dict[str, Any],
        tags: list[dict[str, str]],
    ) -> str:
        """Create a Kinesis Data Firehose delivery stream."""
        (destination_name, destination_config) = find_destination_config_in_args(
            locals()
        )

        if delivery_stream_name in self.delivery_streams:
            raise ResourceInUseException(
                f"Firehose {delivery_stream_name} under accountId {self.account_id} "
                f"already exists"
            )

        if len(self.delivery_streams) == DeliveryStream.MAX_STREAMS_PER_REGION:
            raise LimitExceededException(
                f"You have already consumed your firehose quota of "
                f"{DeliveryStream.MAX_STREAMS_PER_REGION} hoses. Firehose "
                f"names: {list(self.delivery_streams.keys())}"
            )

        # Rule out situations that are not yet implemented.
        if destination_name == "Splunk":
            warnings.warn(
                "A Splunk destination delivery stream is not yet implemented",
                stacklevel=2,
            )

        if (
            kinesis_stream_source_configuration
            and delivery_stream_type != "KinesisStreamAsSource"
        ):
            raise InvalidArgumentException(
                "KinesisSourceStreamConfig is only applicable for "
                "KinesisStreamAsSource stream type"
            )

        # Validate the tags before proceeding.
        errmsg = self.tagger.validate_tags(tags or [])
        if errmsg:
            raise ValidationException(errmsg)

        if tags and len(tags) > MAX_TAGS_PER_DELIVERY_STREAM:
            raise ValidationException(
                f"1 validation error detected: Value '{tags}' at 'tags' "
                f"failed to satisify contstraint: Member must have length "
                f"less than or equal to {MAX_TAGS_PER_DELIVERY_STREAM}"
            )

        # Create a DeliveryStream instance that will be stored and indexed
        # by delivery stream name.  This instance will update the state and
        # create the ARN.
        delivery_stream = DeliveryStream(
            account_id=self.account_id,
            region=region,
            delivery_stream_name=delivery_stream_name,
            delivery_stream_type=delivery_stream_type,
            encryption=delivery_stream_encryption_configuration_input,
            kinesis_stream_source_configuration=kinesis_stream_source_configuration,
            destination_name=destination_name,
            destination_config=destination_config,
        )
        self.tagger.tag_resource(delivery_stream.delivery_stream_arn, tags or [])

        self.delivery_streams[delivery_stream_name] = delivery_stream
        return self.delivery_streams[delivery_stream_name].delivery_stream_arn

    def delete_delivery_stream(self, delivery_stream_name: str) -> None:
        """Delete a delivery stream and its data.

        AllowForceDelete option is ignored as we only superficially
        apply state.
        """
        delivery_stream = self.delivery_streams.get(delivery_stream_name)
        if not delivery_stream:
            raise ResourceNotFoundException(
                f"Firehose {delivery_stream_name} under account {self.account_id} "
                f"not found."
            )

        self.tagger.delete_all_tags_for_resource(delivery_stream.delivery_stream_arn)

        delivery_stream.delivery_stream_status = "DELETING"
        self.delivery_streams.pop(delivery_stream_name)

    def describe_delivery_stream(self, delivery_stream_name: str) -> dict[str, Any]:
        """Return description of specified delivery stream and its status.

        Note:  the 'limit' and 'exclusive_start_destination_id' parameters
        are not currently processed/implemented.
        """
        delivery_stream = self.delivery_streams.get(delivery_stream_name)
        if not delivery_stream:
            raise ResourceNotFoundException(
                f"Firehose {delivery_stream_name} under account {self.account_id} "
                f"not found."
            )

        result: dict[str, Any] = {
            "DeliveryStreamDescription": {"HasMoreDestinations": False}
        }
        for attribute, attribute_value in vars(delivery_stream).items():
            if not attribute_value:
                continue

            # Convert from attribute's snake case to camel case for outgoing
            # JSON.
            name = "".join([x.capitalize() for x in attribute.split("_")])

            # Fooey ... always an exception to the rule:
            if name == "DeliveryStreamArn":
                name = "DeliveryStreamARN"

            if name != "Destinations":
                if name == "Source":
                    result["DeliveryStreamDescription"][name] = {
                        "KinesisStreamSourceDescription": attribute_value
                    }
                else:
                    result["DeliveryStreamDescription"][name] = attribute_value
                continue

            result["DeliveryStreamDescription"]["Destinations"] = []
            for destination in attribute_value:
                description = {}
                for key, value in destination.items():
                    if key == "destination_id":
                        description["DestinationId"] = value
                    else:
                        description[f"{key}DestinationDescription"] = value

                result["DeliveryStreamDescription"]["Destinations"].append(description)

        return result

    def list_delivery_streams(
        self,
        limit: Optional[int],
        delivery_stream_type: str,
        exclusive_start_delivery_stream_name: str,
    ) -> dict[str, Any]:
        """Return list of delivery streams in alphabetic order of names."""
        result = {"DeliveryStreamNames": [], "HasMoreDeliveryStreams": False}
        if not self.delivery_streams:
            return result

        # If delivery_stream_type is specified, filter out any stream that's
        # not of that type.
        stream_list = list(self.delivery_streams.keys())
        if delivery_stream_type:
            stream_list = [
                x
                for x in stream_list
                if self.delivery_streams[x].delivery_stream_type == delivery_stream_type
            ]

        # The list is sorted alphabetically, not alphanumerically.
        sorted_list = sorted(stream_list)

        # Determine the limit or number of names to return in the list.
        limit = limit or DeliveryStream.MAX_STREAMS_PER_REGION

        # If a starting delivery stream name is given, find the index into
        # the sorted list, then add one to get the name following it.  If the
        # exclusive_start_delivery_stream_name doesn't exist, it's ignored.
        start = 0
        if exclusive_start_delivery_stream_name:
            if self.delivery_streams.get(exclusive_start_delivery_stream_name):
                start = sorted_list.index(exclusive_start_delivery_stream_name) + 1

        result["DeliveryStreamNames"] = sorted_list[start : start + limit]
        if len(sorted_list) > (start + limit):
            result["HasMoreDeliveryStreams"] = True
        return result

    def list_tags_for_delivery_stream(
        self,
        delivery_stream_name: str,
        exclusive_start_tag_key: str,
        limit: Optional[int],
    ) -> dict[str, Any]:
        """Return list of tags."""
        result = {"Tags": [], "HasMoreTags": False}
        delivery_stream = self.delivery_streams.get(delivery_stream_name)
        if not delivery_stream:
            raise ResourceNotFoundException(
                f"Firehose {delivery_stream_name} under account {self.account_id} not found."
            )

        tags = self.tagger.list_tags_for_resource(delivery_stream.delivery_stream_arn)[
            "Tags"
        ]
        keys = self.tagger.extract_tag_names(tags)

        # If a starting tag is given and can be found, find the index into
        # tags, then add one to get the tag following it.
        start = 0
        if exclusive_start_tag_key:
            if exclusive_start_tag_key in keys:
                start = keys.index(exclusive_start_tag_key) + 1

        limit = limit or MAX_TAGS_PER_DELIVERY_STREAM
        result["Tags"] = tags[start : start + limit]
        if len(tags) > (start + limit):
            result["HasMoreTags"] = True
        return result

    def put_record(
        self, delivery_stream_name: str, record: dict[str, bytes]
    ) -> dict[str, Any]:
        """Write a single data record into a Kinesis Data firehose stream."""
        result = self.put_record_batch(delivery_stream_name, [record])
        return {
            "RecordId": result["RequestResponses"][0]["RecordId"],
            "Encrypted": False,
        }

    @staticmethod
    def put_http_records(  # type: ignore[misc]
        http_destination: dict[str, Any], records: list[dict[str, bytes]]
    ) -> list[dict[str, str]]:
        """Put records to a HTTP destination."""
        # Mostly copied from localstack
        url = http_destination["EndpointConfiguration"]["Url"]
        headers = {"Content-Type": "application/json"}
        record_to_send = {
            "requestId": str(mock_random.uuid4()),
            "timestamp": int(time()),
            "records": [
                {"data": str(Base64EncodedString.from_raw_bytes(record["Data"]))}
                for record in records
            ],
        }
        try:
            requests.post(url, json=record_to_send, headers=headers)
        except Exception as exc:
            # This could be better ...
            raise RuntimeError(
                "Firehose PutRecord(Batch) to HTTP destination failed"
            ) from exc
        return [{"RecordId": str(mock_random.uuid4())} for _ in range(len(records))]

    @staticmethod
    def _format_s3_object_path(
        delivery_stream_name: str, version_id: str, prefix: str
    ) -> str:
        """Return a S3 object path in the expected format."""
        # Taken from LocalStack's firehose logic, with minor changes.
        # See https://docs.aws.amazon.com/firehose/latest/dev/basic-deliver.html#s3-object-name
        # Path prefix pattern: myApp/YYYY/MM/DD/HH/
        # Object name pattern:
        # DeliveryStreamName-DeliveryStreamVersion-YYYY-MM-DD-HH-MM-SS-RandomString
        prefix = f"{prefix}{'' if prefix.endswith('/') or prefix == '' else '/'}"
        now = utcnow()
        return (
            f"{prefix}{now.strftime('%Y/%m/%d/%H')}/"
            f"{delivery_stream_name}-{version_id}-"
            f"{now.strftime('%Y-%m-%d-%H-%M-%S')}-{str(mock_random.uuid4())}"
        )

    def put_s3_records(
        self,
        delivery_stream_name: str,
        version_id: str,
        s3_destination: dict[str, Any],
        records: list[dict[str, bytes]],
    ) -> list[dict[str, str]]:
        """Put records to a ExtendedS3 or S3 destination."""
        # Taken from LocalStack's firehose logic, with minor changes.
        bucket_name = s3_destination["BucketARN"].split(":")[-1]
        prefix = s3_destination.get("Prefix", "")
        object_path = self._format_s3_object_path(
            delivery_stream_name, version_id, prefix
        )

        batched_data = b"".join([r["Data"] for r in records])
        try:
            s3_backends[self.account_id][self.partition].put_object(
                bucket_name, object_path, batched_data
            )
        except Exception as exc:
            # This could be better ...
            raise RuntimeError(
                "Firehose PutRecord(Batch to S3 destination failed"
            ) from exc
        return [{"RecordId": str(mock_random.uuid4())} for _ in range(len(records))]

    def put_record_batch(
        self, delivery_stream_name: str, records: list[dict[str, bytes]]
    ) -> dict[str, Any]:
        """Write multiple data records into a Kinesis Data firehose stream."""
        delivery_stream = self.delivery_streams.get(delivery_stream_name)
        if not delivery_stream:
            raise ResourceNotFoundException(
                f"Firehose {delivery_stream_name} under account {self.account_id} "
                f"not found."
            )

        request_responses = []
        for destination in delivery_stream.destinations:
            if "ExtendedS3" in destination:
                # ExtendedS3 will be handled like S3,but in the future
                # this will probably need to be revisited.  This destination
                # must be listed before S3 otherwise both destinations will
                # be processed instead of just ExtendedS3.
                request_responses = self.put_s3_records(
                    delivery_stream_name,
                    delivery_stream.version_id,
                    destination["ExtendedS3"],
                    records,
                )
            elif "S3" in destination:
                request_responses = self.put_s3_records(
                    delivery_stream_name,
                    delivery_stream.version_id,
                    destination["S3"],
                    records,
                )
            elif "HttpEndpoint" in destination:
                request_responses = self.put_http_records(
                    destination["HttpEndpoint"], records
                )
            elif {"Elasticsearch", "Redshift", "Snowflake"} & set(destination):
                # This isn't implemented as these services aren't implemented,
                # so ignore the data, but return a "proper" response.
                request_responses = [
                    {"RecordId": str(mock_random.uuid4())} for _ in range(len(records))
                ]

        return {
            "FailedPutCount": 0,
            "Encrypted": False,
            "RequestResponses": request_responses,
        }

    def tag_delivery_stream(
        self, delivery_stream_name: str, tags: list[dict[str, str]]
    ) -> None:
        """Add/update tags for specified delivery stream."""
        delivery_stream = self.delivery_streams.get(delivery_stream_name)
        if not delivery_stream:
            raise ResourceNotFoundException(
                f"Firehose {delivery_stream_name} under account {self.account_id} "
                f"not found."
            )

        if len(tags) > MAX_TAGS_PER_DELIVERY_STREAM:
            raise ValidationException(
                f"1 validation error detected: Value '{tags}' at 'tags' "
                f"failed to satisify contstraint: Member must have length "
                f"less than or equal to {MAX_TAGS_PER_DELIVERY_STREAM}"
            )

        errmsg = self.tagger.validate_tags(tags)
        if errmsg:
            raise ValidationException(errmsg)

        self.tagger.tag_resource(delivery_stream.delivery_stream_arn, tags)

    def untag_delivery_stream(
        self, delivery_stream_name: str, tag_keys: list[str]
    ) -> None:
        """Removes tags from specified delivery stream."""
        delivery_stream = self.delivery_streams.get(delivery_stream_name)
        if not delivery_stream:
            raise ResourceNotFoundException(
                f"Firehose {delivery_stream_name} under account {self.account_id} not found."
            )

        # If a tag key doesn't exist for the stream, boto3 ignores it.
        self.tagger.untag_resource_using_names(
            delivery_stream.delivery_stream_arn, tag_keys
        )

    def start_delivery_stream_encryption(
        self, stream_name: str, encryption_config: dict[str, Any]
    ) -> None:
        delivery_stream = self.delivery_streams.get(stream_name)
        if not delivery_stream:
            raise ResourceNotFoundException(
                f"Firehose {stream_name} under account {self.account_id} not found."
            )

        delivery_stream.delivery_stream_encryption_configuration = encryption_config
        delivery_stream.delivery_stream_encryption_configuration["Status"] = "ENABLED"

    def stop_delivery_stream_encryption(self, stream_name: str) -> None:
        delivery_stream = self.delivery_streams.get(stream_name)
        if not delivery_stream:
            raise ResourceNotFoundException(
                f"Firehose {stream_name} under account {self.account_id} not found."
            )

        delivery_stream.delivery_stream_encryption_configuration["Status"] = "DISABLED"

    def update_destination(
        self,
        delivery_stream_name: str,
        current_delivery_stream_version_id: str,
        destination_id: str,
        s3_destination_update: dict[str, Any],
        extended_s3_destination_update: dict[str, Any],
        s3_backup_mode: str,
        redshift_destination_update: dict[str, Any],
        elasticsearch_destination_update: dict[str, Any],
        splunk_destination_update: dict[str, Any],
        http_endpoint_destination_update: dict[str, Any],
        snowflake_destination_configuration: dict[str, Any],
    ) -> None:
        (dest_name, dest_config) = find_destination_config_in_args(locals())

        delivery_stream = self.delivery_streams.get(delivery_stream_name)
        if not delivery_stream:
            raise ResourceNotFoundException(
                f"Firehose {delivery_stream_name} under accountId {self.account_id} not found."
            )

        if dest_name == "Splunk":
            warnings.warn(
                "A Splunk destination delivery stream is not yet implemented",
                stacklevel=2,
            )

        if delivery_stream.version_id != current_delivery_stream_version_id:
            raise ConcurrentModificationException(
                f"Cannot update firehose: {delivery_stream_name} since the "
                f"current version id: {delivery_stream.version_id} and "
                f"specified version id: {current_delivery_stream_version_id} "
                f"do not match"
            )

        destination = {}
        destination_idx = 0
        for destination in delivery_stream.destinations:
            if destination["destination_id"] == destination_id:
                break
            destination_idx += 1
        else:
            raise InvalidArgumentException("Destination Id {destination_id} not found")

        # Switching between Amazon ES and other services is not supported.
        # For an Amazon ES destination, you can only update to another Amazon
        # ES destination.  Same with HTTP.  Didn't test Splunk.
        if (dest_name == "Elasticsearch" and "Elasticsearch" not in destination) or (
            dest_name == "HttpEndpoint" and "HttpEndpoint" not in destination
        ):
            raise InvalidArgumentException(
                f"Changing the destination type to or from {dest_name} "
                f"is not supported at this time."
            )

        # If this is a different type of destination configuration,
        # the existing configuration is reset first.
        if dest_name in destination:
            delivery_stream.destinations[destination_idx][dest_name].update(dest_config)
        else:
            delivery_stream.destinations[destination_idx] = {
                "destination_id": destination_id,
                dest_name: dest_config,
            }

        # Some names in the Update-request differ from the names used by Describe
        for old, new in DeliveryStream.ALTERNATIVE_FIELD_NAMES:
            if old in dest_config:
                if new not in delivery_stream.destinations[destination_idx][dest_name]:
                    delivery_stream.destinations[destination_idx][dest_name][new] = {}
                delivery_stream.destinations[destination_idx][dest_name][new].update(
                    dest_config[old]
                )

        # Once S3 is updated to an ExtendedS3 destination, both remain in
        # the destination.  That means when one is updated, the other needs
        # to be updated as well.  The problem is that they don't have the
        # same fields.
        if dest_name == "ExtendedS3":
            delivery_stream.destinations[destination_idx]["S3"] = (
                create_s3_destination_config(dest_config)
            )
        elif dest_name == "S3" and "ExtendedS3" in destination:
            destination["ExtendedS3"] = {
                k: v
                for k, v in destination["S3"].items()
                if k in destination["ExtendedS3"]
            }

        # Increment version number and update the timestamp.
        delivery_stream.version_id = str(int(current_delivery_stream_version_id) + 1)
        delivery_stream.last_update_timestamp = utcnow()

        # Unimplemented: processing of the "S3BackupMode" parameter.  Per the
        # documentation:  "You can update a delivery stream to enable Amazon
        # S3 backup if it is disabled.  If backup is enabled, you can't update
        # the delivery stream to disable it."

    def lookup_name_from_arn(self, arn: str) -> Optional[DeliveryStream]:
        """Given an ARN, return the associated delivery stream name."""
        return self.delivery_streams.get(arn.split("/")[-1])

    def send_log_event(
        self,
        delivery_stream_arn: str,
        filter_name: str,
        log_group_name: str,
        log_stream_name: str,
        log_events: list[dict[str, Any]],
    ) -> None:
        """Send log events to a S3 bucket after encoding and gzipping it."""
        data = {
            "logEvents": log_events,
            "logGroup": log_group_name,
            "logStream": log_stream_name,
            "messageType": "DATA_MESSAGE",
            "owner": self.account_id,
            "subscriptionFilters": [filter_name],
        }

        output = io.BytesIO()
        with GzipFile(fileobj=output, mode="w") as fhandle:
            fhandle.write(json.dumps(data, separators=(",", ":")).encode("utf-8"))
        gzipped_payload = output.getvalue()

        delivery_stream: DeliveryStream = self.lookup_name_from_arn(delivery_stream_arn)  # type: ignore[assignment]
        self.put_s3_records(
            delivery_stream.delivery_stream_name,
            delivery_stream.version_id,
            delivery_stream.destinations[0]["S3"],
            [{"Data": gzipped_payload}],
        )


firehose_backends = BackendDict(FirehoseBackend, "firehose")



from .responses import ConnectResponse

url_bases = [
    r"https?://connect\.(.+)\.amazonaws\.com",
]

url_paths = {
    "{0}/analytics-data/instance/(?P<InstanceId>[^/]+)/association$": ConnectResponse.dispatch,
    "{0}/instance/(?P<InstanceId>[^/]+)$": ConnectResponse.dispatch,
    "{0}/instance$": ConnectResponse.dispatch,
    "{0}/tags/(?P<resourceArn>.+)$": ConnectResponse.dispatch,
}



"""Exceptions raised by the account service."""

from moto.core.exceptions import JsonRESTError


class UnknownContactType(JsonRESTError):
    def __init__(self, user_arn: str, action: str):
        message = f"User: {user_arn} is not authorized to perform: account:{action} (You specified an invalid Alternate Contact type.)"
        super().__init__(error_type="AccessDeniedException", message=message)


class UnspecifiedContactType(JsonRESTError):
    def __init__(self) -> None:
        super().__init__(
            error_type="ResourceNotFoundException",
            message="No contact of the inputted alternate contact type found.",
        )



"""Exceptions raised by the connect service."""

from moto.core.exceptions import JsonRESTError


class ResourceNotFoundException(JsonRESTError):
    code = 404

    def __init__(self, message: str) -> None:
        super().__init__("ResourceNotFoundException", message)


class InvalidParameterException(JsonRESTError):
    code = 400

    def __init__(self, message: str) -> None:
        super().__init__("InvalidParameterException", message)



from typing import Any, Optional

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.core.utils import utcnow
from moto.moto_api._internal import mock_random

from .exceptions import ResourceNotFoundException


class Invitation(BaseModel):
    def __init__(self, account_id: str, region_name: str, admin_account_id: str):
        self.account_id = account_id
        self.invitation_id = mock_random.get_random_hex()
        self.invited_at = utcnow()
        self.relationship_status = "Invited"
        self.arn = f"arn:aws:macie2:{region_name}:{admin_account_id}:invitation/{self.invitation_id}"

    def to_json(self) -> dict[str, Any]:
        return {
            "accountId": self.account_id,
            "invitationId": self.invitation_id,
            "invitedAt": self.invited_at.isoformat(),
            "relationshipStatus": self.relationship_status,
        }


class Member(BaseModel):
    def __init__(
        self,
        account_id: str,
        region_name: str,
        admin_account_id: str,
        invitation: Invitation,
    ):
        self.account_id = account_id
        self.relationship_status = "Enabled"
        self.updated_at = utcnow()
        self.arn = (
            f"arn:aws:macie2:{region_name}:{admin_account_id}:member/{self.account_id}"
        )
        self.administrator_account_id = admin_account_id
        self.invited_at = invitation.invited_at

    def to_json(self) -> dict[str, Any]:
        return {
            "accountId": self.account_id,
            "administratorAccountId": self.administrator_account_id,
            "arn": self.arn,
            "email": f"{self.account_id}@example.com",  # Moto does not have access to real emails
            "invitedAt": self.invited_at.isoformat(),
            "masterAccountId": self.administrator_account_id,  # masterAccountId is deprecated but still present
            "relationshipStatus": self.relationship_status,
            "updatedAt": self.updated_at.isoformat(),
        }


class MacieBackend(BaseBackend):
    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.invitations: dict[str, Invitation] = {}
        self.members: dict[str, Member] = {}
        self.administrator_account: Optional[Member] = None
        self.organization_admin_account_id: Optional[str] = None
        self.macie_session: Optional[dict[str, Any]] = {
            "createdAt": utcnow(),
            "findingPublishingFrequency": "FIFTEEN_MINUTES",
            "serviceRole": f"arn:aws:iam::{account_id}:role/aws-service-role/macie.amazonaws.com/AWSServiceRoleForAmazonMacie",
            "status": "ENABLED",
            "updatedAt": utcnow(),
        }

    def create_invitations(self, account_ids: list[str]) -> None:
        for account_id in account_ids:
            invitation = Invitation(
                account_id=account_id,
                region_name=self.region_name,
                admin_account_id=self.account_id,
            )
            self.invitations[account_id] = invitation

    def list_invitations(self) -> list[Invitation]:
        return list(self.invitations.values())

    def decline_invitations(self, account_ids: list[str]) -> None:
        for account_id in account_ids:
            for backend_dict in macie2_backends.values():
                backend = backend_dict.get(self.region_name)
                if backend and account_id in backend.invitations:
                    backend.invitations.pop(account_id)

    def accept_invitation(
        self, administrator_account_id: str, invitation_id: str
    ) -> None:
        # Loop through every account backend to find the matching invitation
        # Invitations can only be sent to the same region
        # If the region does not exist, then the user hasn't used Macie in this account yet, and we can safely skip it
        for backend_dict in macie2_backends.values():
            backend = backend_dict.get(self.region_name)
            if not backend:
                continue

            invitation = backend.invitations.get(self.account_id)
            if invitation and invitation.invitation_id == invitation_id:
                # Create member relationship
                member = Member(
                    account_id=self.account_id,
                    region_name=self.region_name,
                    admin_account_id=administrator_account_id,
                    invitation=invitation,
                )
                backend.members[self.account_id] = member
                self.administrator_account = member
                backend.invitations.pop(self.account_id)
                return

    def list_members(self) -> list[Member]:
        return list(self.members.values())

    def get_administrator_account(self) -> Optional[Member]:
        return self.administrator_account

    def delete_member(self, member_account_id: str) -> None:
        if member_account_id not in self.members:
            raise ResourceNotFoundException(
                "The request failed because the resource doesn't exist."
            )

        self.members.pop(member_account_id)

        for backend_dict in macie2_backends.values():
            backend = backend_dict.get(self.region_name)
            if backend and backend.account_id == member_account_id:
                backend.administrator_account = None
                break

    def disassociate_member(self, member_account_id: str) -> None:
        if member_account_id not in self.members:
            raise ResourceNotFoundException(
                "The request failed because the resource doesn't exist."
            )

        self.members.pop(member_account_id)

        for backend_dict in macie2_backends.values():
            backend = backend_dict.get(self.region_name)
            if backend and backend.account_id == member_account_id:
                backend.administrator_account = None
                break

    def get_macie_session(self) -> dict[str, Any]:
        if not self.macie_session:
            raise ResourceNotFoundException(
                "The request failed because the specified resource doesn't exist."
            )
        return {
            "createdAt": self.macie_session["createdAt"].isoformat(),
            "findingPublishingFrequency": self.macie_session[
                "findingPublishingFrequency"
            ],
            "serviceRole": self.macie_session["serviceRole"],
            "status": self.macie_session["status"],
            "updatedAt": self.macie_session["updatedAt"].isoformat(),
        }

    def enable_macie(
        self,
        finding_publishing_frequency: str = "FIFTEEN_MINUTES",
        status: str = "ENABLED",
    ) -> None:
        now = utcnow()
        self.macie_session = {
            "createdAt": now,
            "findingPublishingFrequency": finding_publishing_frequency,
            "serviceRole": f"arn:aws:iam::{self.account_id}:role/aws-service-role/macie.amazonaws.com/AWSServiceRoleForAmazonMacie",
            "status": status,
            "updatedAt": now,
        }

    def enable_organization_admin_account(self, admin_account_id: str) -> None:
        self.organization_admin_account_id = admin_account_id

    def list_organization_admin_accounts(self) -> list[dict[str, str]]:
        if self.organization_admin_account_id:
            return [
                {
                    "accountId": self.organization_admin_account_id,
                    "status": "ENABLED",
                }
            ]
        return []

    def disable_macie(self) -> None:
        self.invitations.clear()
        self.members.clear()
        self.administrator_account = None
        self.macie_session = None


macie2_backends = BackendDict(MacieBackend, "macie2")



from moto.core.responses import ActionResult, EmptyResult

from ._base_response import EC2BaseResponse


class CustomerGateways(EC2BaseResponse):
    def create_customer_gateway(self) -> ActionResult:
        gateway_type = self._get_param("Type")
        ip_address = self._get_param("IpAddress") or self._get_param("PublicIp")
        bgp_asn = self._get_param("BgpAsn")
        tags = self._parse_tag_specification().get("customer-gateway", {})
        customer_gateway = self.ec2_backend.create_customer_gateway(
            gateway_type, ip_address=ip_address, bgp_asn=bgp_asn, tags=tags
        )
        result = {"CustomerGateway": customer_gateway}
        return ActionResult(result)

    def delete_customer_gateway(self) -> ActionResult:
        customer_gateway_id = self._get_param("CustomerGatewayId")
        self.ec2_backend.delete_customer_gateway(customer_gateway_id)
        return EmptyResult()

    def describe_customer_gateways(self) -> ActionResult:
        self.error_on_dryrun()
        filters = self._filters_from_querystring()
        customer_gateway_ids = self._get_param("CustomerGatewayIds", [])
        customer_gateways = self.ec2_backend.describe_customer_gateways(
            filters, customer_gateway_ids
        )
        result = {"CustomerGateways": customer_gateways}
        return ActionResult(result)



"""TimestreamInfluxDBBackend class with methods for supported APIs."""

from enum import Enum
from typing import Any, Optional

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.utilities.paginator import paginate
from moto.utilities.tagging_service import TaggingService

from .exceptions import (
    ConflictException,
    ResourceNotFoundException,
    ValidationException,
)
from .utils import random_id, validate_name

PAGINATION_MODEL = {
    "list_db_parameter_groups": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,
        "unique_attribute": "id",
    },
    "list_db_clusters": {
        "input_token": "next_token",
        "limit_key": "max_results",
        "limit_default": 100,
        "unique_attribute": "id",
    },
}


class InstanceStatus(str, Enum):
    CREATING = "CREATING"
    AVAILABLE = "AVAILABLE"
    DELETING = "DELETING"
    MODIFYING = "MODIFYING"
    UPDATING = "UPDATING"
    DELETED = "DELETED"
    FAILED = "FAILED"
    UPDATING_DEPLOYMENT_TYPE = "UPDATING_DEPLOYMENT_TYPE"
    UPDATING_INSTANCE_TYPE = "UPDATING_INSTANCE_TYPE"


class NetworkType(str, Enum):
    IPV4 = "IPV4"
    DUAL = "DUAL"


class InstanceType(str, Enum):
    DB_INFLUX_MEDIUM = "db.influx.medium"
    DB_INFLUX_LARGE = "db.influx.large"
    DB_INFLUX_XLARGE = "db.influx.xlarge"
    DB_INFLUX_2XLARGE = "db.influx.2xlarge"
    DB_INFLUX_4XLARGE = "db.influx.4xlarge"
    DB_INFLUX_8XLARGE = "db.influx.8xlarge"
    DB_INFLUX_12XLARGE = "db.influx.12xlarge"
    DB_INFLUX_16XLARGE = "db.influx.16xlarge"


class DBStorageType(str, Enum):
    InfluxIOIncludedT1 = "InfluxIOIncludedT1"
    InfluxIOIncludedT2 = "InfluxIOIncludedT2"
    InfluxIOIncludedT3 = "InfluxIOIncludedT3"


class DeploymentType(str, Enum):
    SINGLE_AZ = "SINGLE_AZ"
    WITH_MULTIAZ_STANDBY = "WITH_MULTIAZ_STANDBY"


class ParameterGroup(BaseModel):
    def __init__(
        self,
        name: str,
        description: Optional[str] = None,
        parameters: Optional[dict[str, Any]] = None,
        region_name: str = "",
        account_id: str = "",
    ):
        self.id = random_id()
        self.name = name
        self.description = description or ""
        self.parameters = parameters or {}
        self.arn = f"arn:aws:timestream-influxdb:{region_name}:{account_id}:db-parameter-group/{self.id}"

    def to_dict(self) -> dict[str, Any]:
        return {
            "id": self.id,
            "name": self.name,
            "arn": self.arn,
            "description": self.description,
            "parameters": self.parameters,
        }

    def to_summary_dict(self) -> dict[str, str]:
        return {
            "id": self.id,
            "name": self.name,
            "arn": self.arn,
            "description": self.description,
        }


class Cluster(BaseModel):
    def __init__(
        self,
        name: str,
        password: str,
        username: Optional[str] = None,
        organization: Optional[str] = None,
        bucket: Optional[str] = None,
        port: Optional[int] = None,
        db_parameter_group_identifier: Optional[str] = None,
        db_instance_type: Optional[str] = None,
        db_storage_type: Optional[str] = None,
        allocated_storage: Optional[int] = None,
        network_type: Optional[str] = None,
        publicly_accessible: Optional[bool] = None,
        vpc_subnet_ids: Optional[list[str]] = None,
        vpc_security_group_ids: Optional[list[str]] = None,
        deployment_type: Optional[str] = None,
        failover_mode: Optional[str] = None,
        log_delivery_configuration: Optional[dict[str, Any]] = None,
        region_name: str = "",
        account_id: str = "",
        endpoint_id: str = "",
    ):
        self.id = random_id()
        self.name = name
        self.password = password
        self.username = username
        self.organization = organization
        self.bucket = bucket
        self.port = port or 8086
        self.db_parameter_group_identifier = db_parameter_group_identifier
        self.db_instance_type = db_instance_type or "db.influx.medium"
        self.db_storage_type = db_storage_type or DBStorageType.InfluxIOIncludedT1
        self.allocated_storage = allocated_storage or 100
        self.network_type = network_type or NetworkType.IPV4
        self.publicly_accessible = publicly_accessible or False
        self.vpc_subnet_ids = vpc_subnet_ids or ["subnet-default"]
        self.vpc_security_group_ids = vpc_security_group_ids or ["sg-default"]
        self.deployment_type = deployment_type or "MULTI_NODE_READ_REPLICAS"
        self.failover_mode = failover_mode or "AUTOMATIC"
        self.log_delivery_configuration = log_delivery_configuration or {}
        self.arn = f"arn:aws:timestream-influxdb:{region_name}:{account_id}:db-cluster/{self.id}"
        self.endpoint = (
            f"{self.id}-{endpoint_id}.timestream-influxdb.{region_name}.on.aws"
        )
        self.reader_endpoint = (
            f"{self.id}-{endpoint_id}.reader.timestream-influxdb.{region_name}.on.aws"
        )
        self.influx_auth_parameters_secret_arn = f"arn:aws:secretsmanager:{region_name}:{account_id}:secret:timestream-influxdb/{self.id}/auth-params-{random_id(6)}"
        self.status = "CREATING"

    def to_dict(self) -> dict[str, Any]:
        return {
            "id": self.id,
            "name": self.name,
            "arn": self.arn,
            "status": self.status,
            "endpoint": self.endpoint,
            "readerEndpoint": self.reader_endpoint,
            "port": self.port,
            "deploymentType": self.deployment_type,
            "dbInstanceType": self.db_instance_type,
            "networkType": self.network_type,
            "dbStorageType": self.db_storage_type,
            "allocatedStorage": self.allocated_storage,
            "publiclyAccessible": self.publicly_accessible,
            "dbParameterGroupIdentifier": self.db_parameter_group_identifier,
            "logDeliveryConfiguration": self.log_delivery_configuration,
            "influxAuthParametersSecretArn": self.influx_auth_parameters_secret_arn,
            "vpcSubnetIds": self.vpc_subnet_ids,
            "vpcSecurityGroupIds": self.vpc_security_group_ids,
            "failoverMode": self.failover_mode,
        }

    def to_summary_dict(self) -> dict[str, Any]:
        return {
            "id": self.id,
            "name": self.name,
            "arn": self.arn,
            "status": self.status,
            "endpoint": self.endpoint,
            "readerEndpoint": self.reader_endpoint,
            "port": self.port,
            "deploymentType": self.deployment_type,
            "dbInstanceType": self.db_instance_type,
            "networkType": self.network_type,
            "dbStorageType": self.db_storage_type,
            "allocatedStorage": self.allocated_storage,
        }


class DBInstance(BaseModel):
    def __init__(
        self,
        name: str,
        username: Optional[str],
        password: str,
        organization: str,
        bucket: str,
        dbInstanceType: str,
        vpcSubnetIds: list[str],
        vpcSecurityGroupIds: list[str],
        publiclyAccessible: bool,
        dbStorageType: str,
        allocatedStorage: int,
        dbParameterGroupIdentifier: Optional[str],
        deploymentType: str,
        logDeliveryConfiguration: Optional[dict[str, Any]],
        tags: Optional[dict[str, Any]],
        port: int,
        networkType: str,
        region_name: str,
        account_id: str,
        endpoint_id: str,
    ):
        # Generate a random id of size 10
        self.id = random_id()

        self.name = name
        self.username = username
        self.password = password
        self.organization = organization
        self.bucket = bucket
        self.db_instance_type = dbInstanceType
        self.vpc_subnet_ids = vpcSubnetIds
        self.vpc_security_group_ids = vpcSecurityGroupIds
        self.publicly_accessible = publiclyAccessible
        self.db_storage_type = dbStorageType
        self.allocated_storage = allocatedStorage
        self.db_parameter_group_id = dbParameterGroupIdentifier
        self.deployment_type = deploymentType
        self.log_delivery_configuration = logDeliveryConfiguration
        self.port = port
        self.network_type = networkType
        self.status = InstanceStatus.CREATING
        self.arn = f"arn:aws:timestream-influxdb:{region_name}:{account_id}:db-instance/{self.id}"
        self.endpoint = (
            f"{self.id}-{endpoint_id}.timestream-influxdb.{region_name}.on.aws"
        )
        # Before 12/09/2024, there was a different endpoint format.
        self.endpoint_old = (
            f"{self.name}-{endpoint_id}.timestream-influxdb.{region_name}.on.aws"
        )

        self.availability_zone = ""  # TODO implement this
        self.secondary_availability_zone = ""  # TODO implement this

    def to_dict(self) -> dict[str, Any]:
        return {
            "id": self.id,
            "name": self.name,
            "arn": self.arn,
            "status": self.status,
            "endpoint": self.endpoint,
            "port": self.port,
            "networkType": self.network_type,
            "dbInstanceType": self.db_instance_type,
            "dbStorageType": self.db_storage_type,
            "allocatedStorage": self.allocated_storage,
            "deploymentType": self.deployment_type,
            "vpcSubnetIds": self.vpc_subnet_ids,
            "publiclyAccessible": self.publicly_accessible,
            "vpcSecurityGroupIds": self.vpc_security_group_ids,
            "dbParameterGroupIdentifier": self.db_parameter_group_id,  # TODO implement this
            "availabilityZone": self.availability_zone,  # TODO implement this
            "secondaryAvailabilityZone": self.secondary_availability_zone,  # TODO implement this
            "logDeliveryConfiguration": self.log_delivery_configuration,  # TODO implement this
            "influxAuthParametersSecretArn": "",  # TODO implement this
        }


class TimestreamInfluxDBBackend(BaseBackend):
    """Implementation of TimestreamInfluxDB APIs."""

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)

        # the endpoint identifier is unique per account and per region
        # https://docs.aws.amazon.com/timestream/latest/developerguide/timestream-for-influxdb.html
        self.endpoint_id: str = random_id(10)
        self.db_instances: dict[str, DBInstance] = {}
        self.db_parameter_groups: dict[str, ParameterGroup] = {}
        self.db_clusters: dict[str, Cluster] = {}
        self.tagger = TaggingService()

    def create_db_instance(
        self,
        name: str,
        username: Optional[str],  # required if using InfluxDB UI though
        password: str,
        organization: str,
        bucket: str,
        db_instance_type: str,
        vpc_subnet_ids: list[str],
        vpc_security_group_ids: list[str],
        db_storage_type: str,
        publicly_accessible: bool,
        allocated_storage: int,
        db_parameter_group_identifier: str,
        deployment_type: str,
        log_delivery_configuration: Optional[dict[str, Any]],
        tags: Optional[dict[str, str]],
        port: int,
        network_type: str,
    ) -> DBInstance:
        """
        dbParameterGroupIdentifier argument is not yet handled
        deploymentType currently is auto set to 'SINGLE_AZ' if not passed in.
        publicAccessible is not yet handled
        logDeliveryConfiguration is not yet handled
        AvailabilityZone and SecondaryAvailabilityZone are not yet handled
        influxAuthParametersSecretArn is not yet handled
        """

        # Checks:
        for db_instance in self.db_instances.values():
            if db_instance.name == name:
                raise ConflictException(
                    f"A DB Instance with the name {name} already exists"
                )

        validate_name(name)

        if db_storage_type not in [t.value for t in DBStorageType]:
            raise ValidationException(f"Unknown DB storage type {db_storage_type}")

        if db_instance_type not in [t.value for t in InstanceType]:
            raise ValidationException(f"Unknown DB instance type {db_instance_type}")

        new_instance = DBInstance(
            name,
            username,
            password,
            organization,
            bucket,
            db_instance_type,
            vpc_subnet_ids,
            vpc_security_group_ids,
            publicly_accessible,
            db_storage_type,
            allocated_storage,
            db_parameter_group_identifier,
            deployment_type,
            log_delivery_configuration,
            tags,
            port,
            network_type,
            self.region_name,
            self.account_id,
            self.endpoint_id,
        )

        # add to the list
        self.db_instances[new_instance.id] = new_instance

        # add tags
        if tags:
            self.tag_resource(new_instance.arn, tags)

        return new_instance

    def delete_db_instance(self, id: str) -> DBInstance:
        if id not in self.db_instances:
            raise ResourceNotFoundException(f"DB Instance with id {id} not found")

        # mark as deleting
        self.db_instances[id].status = InstanceStatus.DELETING
        return self.db_instances.pop(id)

    def get_db_instance(self, id: str) -> DBInstance:
        if id not in self.db_instances:
            raise ResourceNotFoundException(f"DB Instance with id {id} not found")

        return self.db_instances[id]

    def list_db_instances(self) -> list[dict[str, Any]]:
        """
        Pagination is not yet implemented
        """
        return [
            {
                "allocatedStorage": instance.allocated_storage,
                "arn": instance.arn,
                "dbInstanceType": instance.db_instance_type,
                "dbStorageType": instance.db_storage_type,
                "deploymentType": instance.deployment_type,
                "endpoint": instance.endpoint,
                "id": instance.id,
                "name": instance.name,
                "networkType": instance.network_type,
                "port": instance.port,
                "status": instance.status,
            }
            for instance in self.db_instances.values()
        ]

    def tag_resource(self, resource_arn: str, tags: dict[str, str]) -> None:
        tag_list = self.tagger.convert_dict_to_tags_input(tags)
        errmsg = self.tagger.validate_tags(tag_list)
        if errmsg:
            raise ValidationException(errmsg)
        self.tagger.tag_resource(resource_arn, tag_list)

    def untag_resource(self, resource_arn: str, tag_keys: list[str]) -> None:
        self.tagger.untag_resource_using_names(resource_arn, tag_keys)

    def list_tags_for_resource(self, resource_arn: str) -> dict[str, str]:
        return self.tagger.get_tag_dict_for_resource(resource_arn)

    def create_db_parameter_group(
        self,
        name: str,
        description: Optional[str] = None,
        parameters: Optional[dict[str, Any]] = None,
        tags: Optional[dict[str, str]] = None,
    ) -> ParameterGroup:
        validate_name(name)

        for param_group in self.db_parameter_groups.values():
            if param_group.name == name:
                raise ConflictException(
                    f"A DB parameter group with the name {name} already exists"
                )

        param_group = ParameterGroup(
            name=name,
            description=description,
            parameters=parameters,
            region_name=self.region_name,
            account_id=self.account_id,
        )

        self.db_parameter_groups[param_group.id] = param_group

        if tags:
            self.tag_resource(param_group.arn, tags)

        return param_group

    def get_db_parameter_group(self, identifier: str) -> ParameterGroup:
        param_group = self.db_parameter_groups.get(identifier)
        if not param_group:
            raise ResourceNotFoundException(
                f"DB parameter group with identifier {identifier} not found"
            )

        return param_group

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_db_parameter_groups(self) -> list[dict[str, str]]:
        if not self.db_parameter_groups:
            return []

        return [
            param_group.to_summary_dict()
            for param_group in self.db_parameter_groups.values()
        ]

    @paginate(pagination_model=PAGINATION_MODEL)
    def list_db_clusters(self) -> list[dict[str, object]]:
        if not self.db_clusters:
            return []

        return [
            cluster.to_summary_dict()
            for cluster in self.db_clusters.values()
            if cluster.status != "DELETED"
        ]

    def get_db_cluster(self, db_cluster_id: str) -> Cluster:
        cluster = self.db_clusters.get(db_cluster_id)
        if not cluster:
            raise ResourceNotFoundException(
                f"DB cluster with ID {db_cluster_id} not found"
            )

        return cluster

    def create_db_cluster(
        self,
        name: str,
        password: str,
        username: Optional[str] = None,
        organization: Optional[str] = None,
        bucket: Optional[str] = None,
        port: Optional[int] = None,
        db_parameter_group_identifier: Optional[str] = None,
        db_instance_type: Optional[str] = None,
        db_storage_type: Optional[str] = None,
        allocated_storage: Optional[int] = None,
        network_type: Optional[str] = None,
        publicly_accessible: Optional[bool] = None,
        vpc_subnet_ids: Optional[list[str]] = None,
        vpc_security_group_ids: Optional[list[str]] = None,
        deployment_type: Optional[str] = None,
        failover_mode: Optional[str] = None,
        log_delivery_configuration: Optional[dict[str, Any]] = None,
        tags: Optional[dict[str, str]] = None,
    ) -> tuple[str, str]:
        validate_name(name)

        for cluster in self.db_clusters.values():
            if cluster.name == name:
                raise ConflictException(
                    f"A DB cluster with the name {name} already exists"
                )

        new_cluster = Cluster(
            name=name,
            password=password,
            username=username,
            organization=organization,
            bucket=bucket,
            port=port,
            db_parameter_group_identifier=db_parameter_group_identifier,
            db_instance_type=db_instance_type,
            db_storage_type=db_storage_type,
            allocated_storage=allocated_storage,
            network_type=network_type,
            publicly_accessible=publicly_accessible,
            vpc_subnet_ids=vpc_subnet_ids,
            vpc_security_group_ids=vpc_security_group_ids,
            deployment_type=deployment_type,
            failover_mode=failover_mode,
            log_delivery_configuration=log_delivery_configuration,
            region_name=self.region_name,
            account_id=self.account_id,
            endpoint_id=self.endpoint_id,
        )

        new_cluster.status = "AVAILABLE"

        self.db_clusters[new_cluster.id] = new_cluster

        if tags:
            self.tag_resource(new_cluster.arn, tags)

        return (
            new_cluster.id,
            "AVAILABLE",
        )


timestreaminfluxdb_backends = BackendDict(
    TimestreamInfluxDBBackend,
    "timestream-influxdb",
    additional_regions=[
        "us-east-1",
        "us-east-2",
        "us-west-2",
        "eu-central-1",
        "eu-west-1",
        "ap-southeast-2",
        "ap-northeast-1",
    ],
)
